"""
    A package that contains utilities for datetime handling
"""

from ._stopwatch import Stopwatch, with_stopwatch
from ._datetime import (
    ensure_timezone,
    ensure_no_timezone,
    to_datetime,
    format_ms_HHMMSSss,
    last_day_of_month,
    month_iterator,
)
from ._ddate import (
    DDateWeekday,
    ddate,
    DDateSeason,
    maybe_get_discordian_season_and_day,
    yold_by_date,
    maybe_get_discordian_weekday,
    decode,
)
from ._gaussian_easter import calculate_easter_date



import asyncio


async def run_command(command, stdout_callback, stderr_callback, **kwds):
    """
    Runs a shell command asynchronously and uses callbacks for stdout and stderr.

    Args:
        command (str): The shell command to run.
        stdout_callback (callable): Function to handle stdout text. Should accept two arguments:
                                    the process instance and the line of output.
        stderr_callback (callable): Function to handle stderr text. Should accept two arguments:
                                    the process instance and the line of output.
    """
    process = await asyncio.create_subprocess_shell(
        command,
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        **kwds
    )

    async def handle_stream(stream, callback):
        """Handle the stream (stdout or stderr) by calling the provided callback."""
        while True:
            line = await stream.readline()
            if not line:
                break
            await callback(process, line.decode().strip())

    # Create tasks to read from stdout and stderr
    stdout_task = asyncio.create_task(handle_stream(process.stdout, stdout_callback))
    stderr_task = asyncio.create_task(handle_stream(process.stderr, stderr_callback))

    # Wait for the process to complete
    await process.wait()

    # Ensure all tasks are completed
    await stdout_task
    await stderr_task
    return process.returncode


async def simple_run_command(command, **kwds):
    """
    Run a shell command asynchronously and print the output to stdout.

    Args:
        command (str): The shell command to run.
    """
    stdout = ""
    stderr = ""

    async def stdout_callback(process, line):
        nonlocal stdout
        stdout += line + "\n"

    async def stderr_callback(process, line):
        nonlocal stderr
        stderr += line + "\n"

    retval = await run_command(command, stdout_callback, stderr_callback, **kwds)

    return retval, stdout, stderr





import unittest

from ratisbona_utils.random_numbers import lsfr_B4BCD35C_32Bit


class MyTestCase(unittest.TestCase):


    def test_lsfr_B4BCD35C_32Bit_must_yield_known_numbers(self):
        """
        Test that the lsfr_B4BCD35C_32Bit must yield known numbers from

        https://www.analog.com/en/resources/design-notes/random-number-generation-using-lfsr.html

        """
        random = lsfr_B4BCD35C_32Bit(0xB4BCD35C)

        numbers = [next(random) for _ in range(4)]

        hex_numbers = [hex(number) for number in numbers]

        self.assertEqual(['0x5a5e69ae', '0x2d2f34d7', '0xa22b4937', '0xe5a977c7'], hex_numbers)

    # Will run more than 10 Minutes!
    def donttest_period_lsfr_B4BCD35C_32Bit_must_be_2pow32(self):
        """
        Test that the lsfr_B4BCD35C_32Bit must yield known numbers from

        https://www.analog.com/en/resources/design-notes/random-number-generation-using-lfsr.html

        """
        random = lsfr_B4BCD35C_32Bit(0xB4BCD35C)

        first_number = next(random)
        period = 1
        while True:
            number = next(random)
            if number == first_number:
                break
            period += 1
            if period % 1_000_000 == 0:
                print(f"{period:_}")
        self.assertEqual(2**32-1, period)

if __name__ == '__main__':
    unittest.main()



from enum import Enum
from typing import Iterable, Optional

from ratisbona_utils.colors import RGBComponent, RGBColor


class TerminalColor(int, Enum):
    BLACK = 0
    RED = 1
    GREEN = 2
    YELLOW = 3
    BLUE = 4
    MAGENTA = 5
    CYAN = 6
    WHITE = 7
    NONE = 999

    def escape(self, bright=True, background=False) -> str:
        if self == TerminalColor.NONE:
            return "\033[0m"
        return (
            f"\033[{'1' if bright else '0'};{'4' if background else '3'}{self.value}m"
        )


def rgb_terminal_escape(rgb: RGBColor, background=False) -> str:
    r, g, b = rgb
    return f"\033[{'48' if background else '38'};2;{r};{g};{b}m"


def color_block(rgb: RGBColor) -> str:
    """
    Prints a block of text with the specified RGB color as background.

    Args:
       rgb (RgbColor): The RGB color to use.
    """
    return rgb_terminal_escape(rgb, background=True) + " " + TerminalColor.NONE.escape()


def color_text(
    text: str, *, foreground: Optional[RGBColor], background: Optional[RGBColor] = None
) -> str | list[str]:
    """
    Prints text in the specified RGB color.

    Args:
        text (str): The text to print.
        foreground (RGBColor): The RGB color to use for the text. Set to none for "leave unchanged".
        background (RGBColor): The RGB color to use for the background. Set to none for "leave unchanged".
    """

    sequence = ""
    if foreground:
        sequence += rgb_terminal_escape(foreground)
    if background:
        sequence += rgb_terminal_escape(background, background=True)

    last_char_is_newline = text[-1] == "\n"

    results = ""
    for line in text.splitlines():
        if len(results) > 0:
            results += "\n"
        results += sequence + line + TerminalColor.NONE.escape()
    if last_char_is_newline:
        results += "\n"

    return results



from .nullsafe import nullsafe, nullsafe_decorator, saveget, substitute, has_keyvalue
from .monads2 import Maybe, Just, Nothing, Error, UnexpectedNothingError

__all__ = [
    "Maybe",
    "Just",
    "Nothing",
    "Error",
    "UnexpectedNothingError",
    "nullsafe",
    "nullsafe_decorator",
    "saveget",
    "substitute",
    "has_keyvalue",
]



from dataclasses import dataclass
from typing import Callable, TypeVar, ContextManager, Optional

from ratisbona_utils.datetime import format_ms_HHMMSSss, Stopwatch

T = TypeVar("T")
Action = str
Milliseconds = int
Profiler = Callable[[Action, Milliseconds], None]


def print_profiling_info(action: Action, time_milliseconds: Milliseconds):
    """
        A simple profiler, that just prints out the profiling information on the console.

        Args:
            action (Action): the name of the action printed on the console, to fit the Profiler-protocol.
            time_milliseconds (Milliseconds): the time in milliseconds the action took. Argument to fit the Profiler-protocol.

        Returns: None

        Side Effects:
            Prints: "Profiling <action> took HHMMSSss."
    """
    print(f"Profiling: {action} took {format_ms_HHMMSSss(time_milliseconds)}")


def with_profile_printing(action: Action, wrapped_function: Callable[[...], T]) -> Callable[[...], T]:
    """
        Applies print_profiling_info (see there) to the callable as a profiler.

        Args:
            action (Action): the name of the action printed on the console.
            wrapped_function (Callable[[...], T]): the callable to profile.
    """
    return with_profiler(print_profiling_info, action, wrapped_function)


def with_profiler(profiler: Profiler, action: Action, wrapped_function: Callable[[...], T]) -> Callable[[...], T]:
    """
        Executes a function informing a profiler about the action and how long it took to do it.

        Args:
            profiler (Profiler): the profiler to be feed the profiling information into. May be None in which case no profiling is performed.
            action (Action): the name of the action executed. Will be given to the profiler.
            wrapped_function (Callable[[...], T]): the function to be profiled.
    """
    if not profiler:
        return wrapped_function

    def _wrapped(*args, **kwargs) -> T:
        with Stopwatch() as stopwatch:
            result = wrapped_function(*args, **kwargs)
        profiler(action, stopwatch.time_elapsed_millis)
        return result

    return _wrapped


@dataclass
class Profiling(ContextManager):
    """
        A context manager to profile an action with a given profiler.
        For example:
        ```
        with Profiling("My action"):
            do_something()
        ```
        will print:
        ```
        Profiling: My action took 00:00:00.123
        ```
        to the console.

        Args:
            action (Action): The name of the action to profile.
            profiler (Profiler): The profiler to use. Defaults to print_profiling_info.
    """
    action: Action
    profiler: Profiler = print_profiling_info
    _stop_watch: Optional[Stopwatch] = None

    def __enter__(self) -> "Profiling":
        self._stop_watch = Stopwatch()
        self._stop_watch.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stop_watch.stop()
        self.profiler(self.action, self._stop_watch.time_elapsed_millis)


import unittest

from ratisbona_utils.binary import BitStuffer
from ratisbona_utils.binary.binary_tools import BitUnstuffer


class MyTestCase(unittest.TestCase):
    def test_bitstuffer_shoud_stuff_as_much_bits_as_requested(self):
        bit_stuffer = BitStuffer()
        bit_stuffer.stuff_bits(0xFE_CA_12_34, 32)
        bit_stuffer.flush()
        self.assertEqual(bit_stuffer.to_bytes(), b"\xfe\xca\x12\x34")

        bit_stuffer = BitStuffer()
        bit_stuffer.stuff_bits(0xFE_CA_12_34, 16)
        bit_stuffer.flush()
        self.assertEqual(
            bit_stuffer.to_bytes(),
            b"\x12\x34",
        )

        bit_stuffer = BitStuffer()
        bit_stuffer.stuff_bits(0xFE_CA_12_34, 8)
        bit_stuffer.flush()
        self.assertEqual(
            bit_stuffer.to_bytes(),
            b"\x34",
        )

    def test_bitstuffer_must_fill_up_to_next_byte_with_0_on_flush(self):
        bit_stuffer = BitStuffer()
        bit_stuffer.stuff_bits(0xFE_CA_12_34, 15)
        bit_stuffer.flush()
        self.assertEqual(bit_stuffer.to_bytes(), (0x12_34 << 1).to_bytes(2, "big"))

    def test_unstuffer_must_reconstruct_bits(self):
        bit_unstuffer = BitUnstuffer(b"\x12\x34\x56\x78\xab\xcd\xef")
        bits = bit_unstuffer.get_bits(8)
        self.assertEqual(0x12, bits)
        bits = bit_unstuffer.get_bits(8)
        self.assertEqual(bits, 0x34)
        bits = bit_unstuffer.get_bits(8)
        self.assertEqual(bits, 0x56)
        bits = bit_unstuffer.get_bits(8)
        self.assertEqual(bits, 0x78)
        bits = bit_unstuffer.get_bits(8)
        self.assertEqual(bits, 0xAB)
        bits = bit_unstuffer.get_bits(8)
        self.assertEqual(bits, 0xCD)
        bits = bit_unstuffer.get_bits(8)
        self.assertEqual(bits, 0xEF)

    def test_stuff_and_unstuff_bits(self):
        random_testdata_3x_2to33_bits = [
            [2, 2, 3],
            [7, 5, 5],
            [1, 14, 5],
            [2, 11, 25],
            [29, 28, 9],
            [46, 27, 105],
            [53, 253, 251],
            [382, 38, 390],
            [500, 381, 480],
            [930, 761, 81],
            [1660, 535, 1064],
            [908, 3558, 2767],
            [1407, 6494, 12771],
            [25518, 1411, 19081],
            [65443, 23188, 44789],
            [113048, 65580, 30403],
            [188816, 148995, 215689],
            [499048, 300758, 88917],
            [109909, 28447, 1038224],
            [1241379, 754010, 623709],
            [840621, 493003, 2146724],
            [7080578, 6633325, 5866086],
            [13514231, 15110624, 8439533],
            [24330854, 33165880, 6505710],
            [7295787, 55200039, 5178595],
            [28429607, 6099813, 26815720],
            [235411492, 44675367, 47901737],
            [383381405, 339357961, 304902964],
            [471193697, 515116583, 989066118],
            [1064575415, 1010707593, 509209638],
            [3943487889, 3111298040, 3128685376],
        ]
        stuffer = BitStuffer()
        for idx, testdata in enumerate(random_testdata_3x_2to33_bits):
            for value in testdata:
                stuffer.stuff_bits(value, idx+2)
        binary = stuffer.to_bytes()

        unstuffer = BitUnstuffer(binary)
        for idx, testdata in enumerate(random_testdata_3x_2to33_bits):
            print(f'Testing {idx+2} Bits...')
            for expect_value in testdata:
                is_value = unstuffer.get_bits(idx + 2)
                print(f'Expected: {expect_value:x} Got: {is_value:x}')
                self.assertEqual(is_value, expect_value)


if __name__ == "__main__":
    unittest.main()



import inspect
import re
from typing import Callable, Sequence
from unidecode import unidecode
from transliterate import translit
from transliterate.exceptions import LanguageDetectionError

CommandOption = str

cleaners: dict[CommandOption, tuple[Callable[[str], str], str]] = {}

# According to ttps://help.interfaceware.com/v6/windows-reserved-file-names
FORBIDDEN_IN_WINDOWS = r'< > : " / \ | ? *'.split(" ")
FORBIDDEN_IN_WINDOWS_REPLACE_BY = [
    "_lt_",
    "_gt_",
    "_colon_",
    "_quote_",
    "_slash_",
    "_bslash_",
    "_bar_",
    "_qm_",
    "_star_",
]
UMLAUTS = {"ä": "ae", "ö": "oe", "ü": "ue", "ß": "ss", "Ä": "Ae", "Ö": "Oe", "Ü": "Ue"}
MULTIPLE_UNDERSCORES_PATTERN = re.compile(r"_{2,}")
MULTIPLE_POINTS_PATTERN = re.compile(r"\.{2,}")
MULTIPLE_DASHES_PATTERN = re.compile(r"-{2,}")
MULTIPLE_UNDERSCORES_PATTERN = re.compile(r"_{2,}")
MULITPLE_UNDERSCORE_DASH_PATTERN = re.compile(r"(_-)+_?")
MULITPLE_SPACE_PATTERN = re.compile(r" +")
JUNK_BEFORE_SUFFIX_PATTERN = re.compile(r"[-_ ]+(\.[^\.]+)$")


def get_docstring_summary(func):
    """Returns the first paragraph of a function's docstring (up to the first empty line)."""
    doc = inspect.getdoc(func)  # Get the cleaned-up docstring (None if no docstring)
    if not doc:
        return None  # Return None if there's no docstring

    # Split by lines and stop at the first empty line
    lines = doc.split("\n")
    summary_lines = []

    for line in lines:
        if line.strip() == "":
            break  # Stop at the first empty line
        summary_lines.append(line)

    return " ".join(summary_lines)  # Join lines into a single string


def _register_cleaner(function: Callable[[str], str]):
    name = function.__name__.replace("sclean_", "")
    docsummary = get_docstring_summary(function)

    cleaners[name] = (function, docsummary)


def sclean_transliterate(current_string: str) -> str:
    """
    Transliterate the string to a latin alphabet representation. This is useful for example for filenames that should
    not contain special characters.

    Args:
        current_string (str): The string to transliterate

    Returns:
        str: The transliterated string
    """
    try:
        return translit(current_string, reversed=True)
    except LanguageDetectionError:
        return current_string


_register_cleaner(sclean_transliterate)


def sclean_unidecode(current_string: str) -> str:
    """
    Unidecode the string to a latin alphabet representation. This is useful for example for filenames that should
    not contain special characters.

    Args:
        current_string (str): The string to unidecode

    Returns:
        str: The unidecoded string

    """
    return unidecode(current_string)


_register_cleaner(sclean_unidecode)


def sclean_tolower(current_string: str) -> str:
    """
    Convert the string to lowercase.

    Args:
        current_string (str): The string to convert

    Returns:
        str: The lowercase string
    """
    return current_string.lower()


_register_cleaner(sclean_tolower)


def sclean_samba_conform(current_string: str) -> str:
    """
    Replace all glyphs that windows won't like in filenames, like < > : " / \\ | ? *

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    for forbidden, replacement in zip(
            FORBIDDEN_IN_WINDOWS, FORBIDDEN_IN_WINDOWS_REPLACE_BY
    ):
        current_string = current_string.replace(forbidden, replacement)
    return current_string


_register_cleaner(sclean_samba_conform)


def sclean_umlauts(current_string: str) -> str:
    """
    Replace all umlauts with their latin alphabet representation, like äöüÄÖÜß -> ae oe ue Ae Oe Ue ss

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    for umlaut, replacement in UMLAUTS.items():
        current_string = current_string.replace(umlaut, replacement)
    return current_string


_register_cleaner(sclean_umlauts)


def sclean_despace(current_string: str) -> str:
    """
    Replace all spaces with underscores

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return current_string.replace(" ", "_")


_register_cleaner(sclean_despace)


def sclean_escape(current_string: str) -> str:
    """
    Replace all escape sequences with '_esc_'

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return current_string.replace("\x1b", "_esc_")


_register_cleaner(sclean_escape)


def sclean_parentesis(current_string: str) -> str:
    """
    Remove all parentesis

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return current_string.replace("(", "").replace(")", "")


_register_cleaner(sclean_parentesis)


def sclean_brackets(current_string: str) -> str:
    """
    Remove all brackets

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return current_string.replace("[", "").replace("]", "")


_register_cleaner(sclean_brackets)


def sclean_curly_brackets(current_string: str) -> str:
    """
    Remove all curly brackets

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return current_string.replace("{", "").replace("}", "")


_register_cleaner(sclean_curly_brackets)


def sclean_leading_space(current_string: str) -> str:
    """
    Remove leading spaces

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return current_string.lstrip()


_register_cleaner(sclean_leading_space)


def sclean_trailing_space(current_string: str) -> str:
    """
    Remove trailing spaces

    Args:
        current_string (str): The string to convert
    Returns:
        str: The cleaned string
    """
    return current_string.rstrip()


_register_cleaner(sclean_trailing_space)


def sclean_leading_dash(current_string: str) -> str:
    """
    Remove leading dashes

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return current_string.lstrip("-")


_register_cleaner(sclean_leading_dash)


def sclean_trailing_dash(current_string: str) -> str:
    """
    Remove trailing dashes

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return current_string.rstrip("-")


_register_cleaner(sclean_trailing_dash)


def sclean_leading_underscore(current_string: str) -> str:
    """
    Remove leading underscores

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return current_string.lstrip("_")


_register_cleaner(sclean_leading_underscore)


def sclean_trailing_underscore(current_string: str) -> str:
    """
    Remove trailing underscores

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return current_string.rstrip("_")


_register_cleaner(sclean_trailing_underscore)


def sclean_multiple_underscores(current_string: str) -> str:
    """
    Replace multiple underscores with a single underscore

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return MULTIPLE_UNDERSCORES_PATTERN.sub("_", current_string)


_register_cleaner(sclean_multiple_underscores)


def sclean_commas(current_string: str) -> str:
    """
    Remove all commas

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return current_string.replace(",", "")


_register_cleaner(sclean_commas)


def sclean_exclamation_marks(current_string: str) -> str:
    """
    Remove all exclamation marks

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return current_string.replace("!", "")


_register_cleaner(sclean_exclamation_marks)


def sclean_single_quotes(current_string: str) -> str:
    """
    Remove all single quotes

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return current_string.replace("'", "")


_register_cleaner(sclean_single_quotes)


def sclean_multiple_points(current_string: str) -> str:
    """
    Replace multiple points with a single point

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return MULTIPLE_POINTS_PATTERN.sub(".", current_string)


_register_cleaner(sclean_multiple_points)


def sclean_ampers_and(current_string: str) -> str:
    """
    Replace ampersands with 'and'

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return current_string.replace("&", "and")


_register_cleaner(sclean_ampers_and)


def sclean_multiple_dashes(current_string: str) -> str:
    """
    Replace multiple dashes with a single dash

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return MULTIPLE_DASHES_PATTERN.sub("-", current_string)


_register_cleaner(sclean_multiple_dashes)


def sclean_multiple_underscores(current_string: str) -> str:
    """
    Replace multiple underscores with a single _-_

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return MULTIPLE_UNDERSCORES_PATTERN.sub("_-_", current_string)


_register_cleaner(sclean_multiple_underscores)


def sclean_multiple_underscores_dashes(current_string: str) -> str:
    """
    Replace multiple underscores with a single _-_

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return MULITPLE_UNDERSCORE_DASH_PATTERN.sub("_-_", current_string)


_register_cleaner(sclean_multiple_underscores_dashes)


def sclean_multiple_spaces(current_string: str) -> str:
    """
    Replace multiple spaces with a single space

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    return MULITPLE_SPACE_PATTERN.sub(" ", current_string)


_register_cleaner(sclean_multiple_spaces)


def sclean_dashes_before_suffix(current_string: str) -> str:
    """
    Remove dashes, whitespace or underscores before a suffix

    Args:
        current_string (str): The string to convert

    Returns:
        str: The cleaned string
    """
    while match := JUNK_BEFORE_SUFFIX_PATTERN.search(current_string):
        current_string = current_string.replace(match.group(0), match.group(1))
    return current_string


_register_cleaner(sclean_dashes_before_suffix)


def string_cleaner(
        current_string: str,
        apply_cleaners: Sequence[str],
        change_callback: Callable[[str, str, str], None] = None,
) -> str:
    """
    Clean a string according to a list of cleaners.

    Args:
        current_string (str): The string to clean
        apply_cleaners (Sequence[str]): The cleaners to apply
        change_callback (Callable[[str, str, str], None], optional): A callback to call when a cleaner changes the string. Defaults to None.
    """
    for cleaner in apply_cleaners:
        if not cleaner in cleaners:
            raise ValueError(f"Unknown cleaner {cleaner}")

    for cleaner in apply_cleaners:
        new_string = cleaners[cleaner][0](current_string)
        if change_callback and new_string != current_string:
            change_callback(cleaner, current_string, new_string)
        current_string = new_string
    return current_string


def main():
    print("Available cleaners:")
    for cleaner, doc in cleaners.items():
        print(f"{cleaner:35s}: {doc[1]}")

if __name__ == "__main__":
    items = list(globals().items())
    for a_item in items:
        if callable(a_item[1]):
            print(a_item[0])
    #main()


from ratisbona_utils.colors import normalized_to_rgb, Palette
from ratisbona_utils.colors.palette_generators import (
    ega_num_to_rgb,
    cga_num_to_normalized_rgb,
)
from ratisbona_utils.colors.simple_color import hex_to_rgb, lab_lighten_rgb

ULIS_WEBCOLORS_HEX = [
    "#993300",
    "#CC0000",
    "#FF6600",
    "#FFCC00",
    "#99CC00",
    "#009900",
    "#00CCCC",
    "#0099FF",
    "#0000CC",
    "#6600CC",
    "#CC00CC",
]

ULIS_WEBCOLORS_RGB = [hex_to_rgb(hexcolor) for hexcolor in ULIS_WEBCOLORS_HEX]
ULIS_WEBCOLORS_PALETTE = Palette(
    name="Ulis Webcolors",
    description="The webcolors of Ulrich",
    author="Ulrich",
    colors=ULIS_WEBCOLORS_RGB,
    color_names=[
        "brown",
        "red",
        "orange",
        "yellow",
        "lime",
        "green",
        "cyan",
        "lightblue",
        "blue",
        "violet",
        "magenta",
    ],
    color_types=["rgb"] * len(ULIS_WEBCOLORS_RGB),
)

FIRE_PALETTE_HEX = [
    "#0000cc",
    "#3333ff",
    "#33ffff",
    #'#330033',
    "#ffff00",
    "#ffff33",
    "#ffcc00",
    "#ff0000",
    "#cc0000",
    "#660000",
    "#000000",
]

FIRE_PALETTE_RGB = [hex_to_rgb(hexcolor) for hexcolor in reversed(FIRE_PALETTE_HEX)]
FIRE_PALETTE = Palette(
    name="Fire",
    description="A fire palette",
    author="Ulrich",
    colors=FIRE_PALETTE_RGB,
    color_names=[
        "blue",
        "lightblue",
        "cyan",
        #'purple',
        "yellow",
        "brightyellow",
        "orange",
        "red",
        "darkred",
        "darkdarkred",
        "black",
    ],
    color_types=["rgb"] * len(FIRE_PALETTE_RGB),
)

EGA_PALETTE_RGB = [ega_num_to_rgb(i) for i in range(64)]
EGA_PALETTE = Palette(
    name="EGA",
    description="The EGA palette",
    author="IBM",
    colors=EGA_PALETTE_RGB,
    color_names=[
        "black",
        "blue",
        "green",
        "cyan",
        "red",
        "magenta",
        "brown",
        "lightgray",
        "darkgray",
        "lightblue",
        "lightgreen",
        "lightcyan",
        "lightred",
        "lightmagenta",
        "yellow",
        "white",
    ],
    color_types=["rgb"] * len(EGA_PALETTE_RGB),
)

CGA_PALETTE_RGB = [normalized_to_rgb(cga_num_to_normalized_rgb(i)) for i in range(16)]
CGA_PALETTE = Palette(
    name="CGA",
    description="The CGA palette",
    author="IBM",
    colors=CGA_PALETTE_RGB,
    color_names=[
        "black",
        "darkblue",
        "darkgreen",
        "darkcyan",
        "darkred",
        "darkmagenta",
        "brown",
        "lightgray",
        "darkgray",
        "lightblue",
        "lightgreen",
        "lightcyan",
        "lightred",
        "lightmagenta",
        "yellow",
        "white",
    ],
    color_types=["rgb"] * len(CGA_PALETTE_RGB),
)

AMIGA_PALETTE_RGB = [(0, 85, 170), (255, 255, 255), (0, 0, 0), (255, 136, 0)]
AMIGA_PALETTE = Palette(
    name="Amiga",
    description="The Amiga palette of Workbench 1.x",
    author="Commodore",
    colors=AMIGA_PALETTE_RGB,
    color_names=["blue", "white", "black", "orange"],
    color_types=["rgb"] * len(AMIGA_PALETTE_RGB),
)

solarized_base03 = (0x0, 0x2B, 0x36)
solarized_base02 = (0x7, 0x36, 0x42)
solarized_base01 = (0x58, 0x6E, 0x75)
solarized_base00 = (0x65, 0x7B, 0x83)
solarized_base0 = (0x83, 0x94, 0x96)
solarized_base1 = (0x93, 0xA1, 0xA1)
solarized_base2 = (0xEE, 0xE8, 0xD5)
solarized_base3 = (0xFD, 0xF6, 0xE3)
solarized_yellow = (0xB5, 0x89, 0x0)
solarized_orange = (0xCB, 0x4B, 0x16)
solarized_red = (0xDC, 0x32, 0x2F)
solarized_magenta = (0xD3, 0x36, 0x82)
solarized_violet = (0x6C, 0x71, 0xC4)
solarized_blue = (0x26, 0x8B, 0xD2)
solarized_cyan = (0x2A, 0xA1, 0x98)
solarized_green = (0x85, 0x99, 0x0)

SOLARIZED_COLORS_BASE = [
    solarized_base03,
    solarized_base02,
    solarized_base01,
    solarized_base00,
    solarized_base0,
    solarized_base1,
    solarized_base2,
    solarized_base3,
]

SOLARIZED_COLORS_COLORS = [
    solarized_red,
    solarized_orange,
    solarized_yellow,
    solarized_green,
    solarized_cyan,
    solarized_blue,
    solarized_violet,
    solarized_magenta,
]

SOLARIZED_COLORS = SOLARIZED_COLORS_BASE + SOLARIZED_COLORS_COLORS
SOLARIZED_COLORS_PALETTE = Palette(
    name="Solarized",
    description="The Solarized palette",
    author="Ethan Schoonover",
    colors=SOLARIZED_COLORS,
    color_names=[
        "base03",
        "base02",
        "base01",
        "base00",
        "base0",
        "base1",
        "base2",
        "base3",
        "red",
        "orange",
        "yellow",
        "green",
        "cyan",
        "blue",
        "violet",
        "magenta",
    ],
    color_types=["rgb"] * len(SOLARIZED_COLORS),
)

SOLARIZED_COLORS_EXTENDED = (
    SOLARIZED_COLORS
    + [lab_lighten_rgb(color, 20) for color in SOLARIZED_COLORS_COLORS]
    + [lab_lighten_rgb(color, -20) for color in SOLARIZED_COLORS_COLORS]
)
SOLARIZED_COLORS_EXTENDED_PALETTE = Palette(
    name="Solarized Extended",
    description="The Solarized palette with lightened and darkened colors",
    author="Ulrich Schwenk, based on work by Ethan Schoonover",
    colors=SOLARIZED_COLORS_EXTENDED,
    color_names=[
        "base03",
        "base02",
        "base01",
        "base00",
        "base0",
        "base1",
        "base2",
        "base3",
        "red",
        "orange",
        "yellow",
        "green",
        "cyan",
        "blue",
        "violet",
        "magenta",
        "red_light",
        "orange_light",
        "yellow_light",
        "green_light",
        "cyan_light",
        "blue_light",
        "violet_light",
        "magenta_light",
        "red_dark",
        "orange_dark",
        "yellow_dark",
        "green_dark",
        "cyan_dark",
        "blue_dark",
        "violet_dark",
        "magenta_dark",
    ],
    color_types=["rgb"] * len(SOLARIZED_COLORS_EXTENDED),
)

SOLARIZED_COLORS_ANSI_TERMINAL_NAMES = [
    "Background", "Black",         "Red",          "Green",        "Yellow",         "Blue",         "Magenta",         "Cyan",         "White",
    "Foreground", "Bright Black",  "Bright Red",   "Bright Green", "Bright Yellow",  "Bright Blue",  "Bright Magenta",  "Bright Cyan",  "Bright White"
]

SOLARIZED_COLORS_ANSI_TERMINAL_HEX = [
    "#163540",     "#0d2a34",       "#ca4238",     "#88982d",       "#ae8a2c",        "#4689cc",      "#c24380",         "#519e97",      "#fbf6e4",
    "#869395",     "#163540",       "#d2736e",     "#bfcb68",       "#e7c24f",        "#6faee8",      "#db73a7",         "#83d0ca",      "#fbf6e4",
]

SOLARIZED_COLORS_ANSI_TERMINAL_PALETTE = Palette(
    name="Solarized Palette for ANSI Terminal",
    description="The Solarized dark palette adapted for use in ANSI terminals",
    author="Ulrich Schwenk, based on work by Ethan Schoonover",
    colors=[hex_to_rgb(color) for color in  SOLARIZED_COLORS_ANSI_TERMINAL_HEX],
    color_names=SOLARIZED_COLORS_ANSI_TERMINAL_NAMES,
    color_types=["rgb"] * len(SOLARIZED_COLORS_ANSI_TERMINAL_HEX),
)


from __future__ import annotations

import sys
import traceback
from typing import TypeVar, Generic, Callable

T = TypeVar("T")
E = TypeVar("E", bound=Exception)
V = TypeVar("V")
U = TypeVar("U")


class UnexpectedNothingError(Exception):
    """
    This error is raised when a value was expected but nothing was encountered,
    for example by the "or_blow_up"-Method.

    Args:
        message (str): The error message.
        *args: The other arguments.

    Fields:
        message (str): The error message.
        traceback (str): The traceback of the error.
        *args: The other arguments
    """

    def __init__(self, *args):
        super().__init__(
            "Nothing was encountered where an value was expected in a Maybe", *args
        )


def key_error_to_nothing(error: Exception) -> Maybe[Exception | None]:
    """
    Converts a KeyError to a Nothing.

    Args:
        error (Exception): The error to convert.

    Returns:
        Maybe[Exception | None]: The converted error.
    """
    if isinstance(error, KeyError):
        return Nothing
    return error


class Maybe(Generic[T]):
    """
    This Maybe Monad can be used as a MaybeMonad or even as an ErrorMonad.

    If you wrap a value that is None, the Maybe will be a Nothing.
    If you wrap a value that inherits from Exception, the Maybe will be an Error.
    If you wrap another Maybe that Maybe will be unwrapped transparently. (If it's
    error or nothing this will be error or nothing, so there is not Just(Nothing) or Just(Error))
    If you finally wrap just an ordinary value then the Maybe will be a Just.
    """

    def __init__(self, value: T | None):
        # Maybes are transparently unwrapped.
        if value is not None and isinstance(value, Maybe):
            self._value = value._value
        else:
            self._value = value

    def __getitem__(self, item):
        """
        Allows to access the value of the monad directly.
        """
        if not self:
            return self

        if isinstance(item, Maybe):
            if not item:
                return item
            item = item._value

        if not hasattr(self._value, "__getitem__"):
            return Error(TypeError(f"{self._value} is not subscriptable"))

        if not item in self._value:
            return Nothing

        return Maybe(self._value[item])

    def __str__(self):
        """
        Returns a string representation of the monad.
        """
        if self.is_just:
            return f"Just({self._value})"
        if self.is_error:
            return f"Error({self._value})"
        return "Nothing"

    def __add__(self, other):
        """
        Adds the value of the monad to another value.
        """

        if isinstance(other, str):
            return str(self) + other

        return NotImplemented

    def __radd__(self, other):
        """
        Adds the value of the monad to another value.
        """

        if isinstance(other, str):
            return other + str(self)

        return NotImplemented

    def __bool__(self):
        """
        Returns True if the Maybe is a Just.
        """
        return self.is_just

    def __eq__(self, other):
        """
        Compares the value of the monad with another value.

        Args:
            other: The other value to compare with.

        Returns:
            bool: True if the value of the monad is equal to the other value, False otherwise.
        """

        if not isinstance(other, Maybe):
            return self._value == other

        return self._value == other._value

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self._value)

    def __iter__(self):
        """
        Returns an iterator over the value of the monad.
        """
        if not self:
            return iter([])

        if not hasattr(self._value, "__iter__"):
            return iter([self])

        return iter([Just(item) for item in self._value])

    def __in__(self, other):
        """
        Returns True if the value of the monad is in the other value.
        """
        if not self:
            return False

        return self._value in other

    def __contains__(self, item):
        """
        Returns True if the value of the monad contains the item.
        """
        if not self:
            return False

        if isinstance(item, Maybe):
            if not item:
                return False
            item = item._value

        return item in self._value or item == self._value

    @staticmethod
    def just(value: T | Maybe[T]) -> Maybe[T]:
        """
        Creates a Just, ensuring that is not Nothing or Error.
        """
        if value is None:
            raise ValueError(
                "None is not what I would call Just a value.\n"
                "I'm assuming you f-ed up, but if you really want a None-Value use the constant Nothing.\n"
                "Finally Maybe(...) will construct a Maybe from a value, if the value is None then\n"
                "Nothing is constructed.\n"
            )

        if isinstance(value, Exception):
            raise ValueError(
                "Well that is an Error and not just a value.\n"
                "I'm assuming you f-ed up, but if you really want an Exception-Value use Error(...) to construct an erorr.\n"
                "Finally Maybe(...) will construct a Maybe from a value, if the value is an Exception then\n"
                "Error is constructed.\n"
            )

        if isinstance(value, Maybe):
            if not value:
                raise ValueError(
                    "You may try to wrap a Just into a Just, in which case you'll end up with\n"
                    "just a Just.\n"
                    "But if you wrap a Nothing or an Error in a Just, I will refuse to do so,\n"
                    "assuming you f-ed up.\n"
                    "If you want to create a Maybe from a Maybe then use Maybe(...) instead,\n"
                    "which will also just return your Maybe.\n"
                )
        return Maybe(value)

    @staticmethod
    def error(value: E | Maybe[E]) -> Maybe[E]:
        """
        Creates an Error, ensuring that is not Nothing or Just.
        """
        if value is None:
            raise ValueError(
                "None is not what I would call an Error.\n"
                "I'm assuming you f-ed up, but if you really want a None-Value use the constant Nothing.\n"
                "Finally Maybe(...) will construct a Maybe from a value, if the value is None then\n"
                "Nothing is constructed.\n"
            )

        if isinstance(value, Maybe):
            if not value.is_error:
                raise ValueError(
                    "You may try to wrap an Error into an Error, in which case you'll end up with\n"
                    "just an Error.\n"
                    "But if you wrap a Nothing or a Just in an Error, I will refuse to do so,\n"
                    "assuming you f-ed up.\n"
                    "If you want to create a Maybe from a Maybe then use Maybe(...) instead,\n"
                    "which will also just return your Maybe.\n"
                )
            return Maybe(value.unwrap_error())

        if not isinstance(value, Exception):
            raise ValueError(
                "Well that is not an Error but a value.\n"
                "I'm assuming you f-ed up, but if you really want a value use Just(...) to construct a value.\n"
                "Finally Maybe(...) will construct a Maybe from a value, if the value is an Exception then\n"
                "Error is constructed.\n"
            )

        return Maybe(value)

    @staticmethod
    def with_errorhandling(
        a_function: Callable[[...], U] | Maybe[Callable[[...], U]], *args, **kwargs
    ) -> Maybe[V]:
        """
        Calls a function with monadic error handling.

        Args:
            a_function (Callable | Maybe[Callable]): The function to call. Will be passed the other arguments.
            The function is allowed to be a Maybe[Callable], in which case the function is unwrapped before calling.
            args (Any | Maybe[Any]): The other arguments to pass to the function
            kwargs (dict[str, Any | Maybe[Any]): The other keyword arguments to pass to the function

        Returns:
            Maybe[V]: The result of the function call, an error or nothing. Surely returns Nothing if function is a Nothing.
            If the function is an error, that error. If any of the arguments is a Nothing or a Maybe, that is returned.
            If the function returns an Error or Nothing, that is returned. If the function throws an error, that is returned.
            If the function returns just a value, than Just(value) is returned.
        """
        DEBUG=False

        if isinstance(a_function, Maybe):
            if not a_function:
                return a_function
            a_function = a_function._value

        unwrapped_args = []
        for arg in args:
            if isinstance(arg, Maybe):
                if not arg:
                    return arg
                arg = arg._value
            unwrapped_args.append(arg)

        unwrapped_kwargs = {}
        for key, value in kwargs.items():
            if isinstance(value, Maybe):
                if not value:
                    return value
                value = value._value
            unwrapped_kwargs[key] = value

        try:
            if DEBUG:
                print(f"About to invoke {a_function.__name__}:")
                for arg in unwrapped_args:
                    print(f"  {arg} ({type(arg)})")
                for key, value in unwrapped_kwargs.items():
                    print(f"  {key}={value} ({type(value)})")

            return Maybe(  # Note: Function might return a Maybe, in which case it will be transparently unwrapped.
                a_function(*unwrapped_args, **unwrapped_kwargs)
            )
        except Exception as any_exception:
            any_exception.add_note(traceback.format_exc())
            if DEBUG:
                print("Caught an exception:", any_exception)
                sys.stdout.flush()
            return Maybe.error(any_exception)

    @property
    def is_just(self):
        """
        Returns True if the Maybe is a Just.
        """
        return self._value is not None and not isinstance(self._value, Exception)

    @property
    def is_error(self):
        """
        Returns True if the Maybe is an Error.
        """
        return isinstance(self._value, Exception)

    @property
    def is_nothing(self):
        """
        Returns True if the Maybe is a Nothing.
        """
        return self._value is None

    def bind(
        self, a_function: Callable[[...], V | Maybe[V]], *args, **kwargs
    ) -> Maybe[V]:
        """
        Binds a function to the ResultMonad. Function can throw exceptions which will
        be caught and returned as an error status in the ResultMonad, for convenience.

        Args:
            a_function (Callable | Maybe[Callable]):
            The function to bind to the result monad. Can be a Maybe[Callable] itself,
            in which case the function is unwrapped before calling.

            args: The other arguments to pass to the function. Maybes can be among them,
            in which case they will be unwrapped before passing.

            kwargs: The other keyword arguments to pass to the function, also maybe unwrapped.

        Returns:
            Maybe[V]: The result of the function call, an error or nothing. If this is nothing, surely nothing.
            If this is error, that error. If there's all right with this but someting fishy with a maybe function,
            that error. If there are errors or nothings in the args, that errors. Finally, if the function throws an
            error, then that error.
        """

        if not self:
            return self

        return Maybe.with_errorhandling(a_function, self, *args, **kwargs)

    def bind_or_bust(
        self, a_function: Callable[[...], V | Maybe[V]], *args, **kwargs
    ) -> V:
        """
        Binds a function to the ResultMonad. Function can throw exceptions which will
        be caught and returned as an error status in the ResultMonad, for convenience.

        Args:
            a_function (Callable | Maybe[Callable]):
            The function to bind to the result monad. Can be a Maybe[Callable] itself,
            in which case the function is unwrapped before calling.

            args: The other arguments to pass to the function. Maybes can be among them,
            in which case they will be unwrapped before passing.

            kwargs: The other keyword arguments to pass to the function, also maybe unwrapped.

        Returns:
            Maybe[V]: The result of the function call, an error or nothing. If this is nothing, surely nothing.
            If this is error, that error. If there's all right with this but someting fishy with a maybe function,
            that error. If there are errors or nothings in the args, that errors. Finally, if the function throws an
            error, then that error.
        """

        return self.bind(a_function, *args, **kwargs).maybe_raise_error()

    def maybe_recover(
        self,
        a_function: Callable[[E | None], T | Maybe[T]],
        *args,
        **kwargs,
    ) -> Maybe[T]:
        """
        Recover from an error or nothing status by applying a function to the Maybe.

        If all is well, function is not called.

        Args:y
            a_function: The function to apply to the error value. Will be passsed the error_value of the
            monad as the first argument or None, if the monad is a Nothing. The function can
            be a Maybe[Callable], in which case the function is unwrapped before calling.
            If the function is an error or a Nothing, the original Error will result.

            args: The other arguments to pass to the function. Will be passed after the error_value. If
            an arg is a Maybe, it will be unwrapped before passing. If an arg is an error or a Nothing,
            the original Error will result.
        """
        if self.is_just:
            return self

        maybe_better = Maybe.with_errorhandling(
            a_function, self._value, *args, **kwargs
        )

        if maybe_better.is_error:
            return self

        return maybe_better

    def maybe_warn(self, message: str | Callable[[T], None]) -> Maybe[T]:
        if not self.is_error:
            return self

        if isinstance(message, str):
            print(message, self._value)

        if isinstance(message, Callable):
            message(self._value)

        return self

    def __or__(self, other: Maybe[V]) -> Maybe[T | V]:
        """
        Return the value of the monad or the value of another monad if the monad is error.
        """
        if not self:
            return other
        return self

    def unwrap_value(self) -> T:
        """
        Get the value of the monad. If the monad is Nothing, None will be returned.

        Returns:
            The value of the monad.

        Raises:
            ValueError: If the monad is an error.
        """
        if self.is_error:
            raise ValueError("You must not unwrap a ResultMonad with an error status.")

        if self.is_nothing:
            raise ValueError("You must not unwrap a ResultMonad with a nothing status.")

        return self._value

    def unwrap_error(self) -> E:
        """
        Get the error value of the monad.

        Returns:
            The error value of the monad.

        Raises:
            ValueError: If the monad is not an error.
        """
        if not self.is_error:
            raise ValueError(
                "You must not unwrap the error of a ResultMonad that in fact is a success."
            )
        return self._value

    def maybe_raise_error(self):
        """
        Raises the error of the monad, if it is an Error.

        Returns:
            The monad, unchanged. Note that if it was an error, well, good luck obtaining the return value...
        """
        if self.is_error:
            raise self._value

        return self

    def value_or_throw(self):
        """
        Unwraps the value of the monad, if it is a Just.
        Raises the error of the monad, if it is an Error.
        Raises UnexpectedNothingError, if the monad is a Nothing.
        """
        self.maybe_raise_error()

        if self.is_nothing:
            unexpected_nothing_error = UnexpectedNothingError()
            unexpected_nothing_error.add_note(
                "Traceback:\n" + "".join(traceback.format_stack())
            )
            raise unexpected_nothing_error

        return self._value

    def default_or_throw(self, default: T) -> T:
        """
        Unwraps the value of the monad, if it is a Just.
        Raises the error of the monad, if it is an Error.
        Returns the default value, if the monad is a Nothing.
        """
        self.maybe_raise_error()

        if self.is_nothing:
            return default

        return self._value

    def default_also_on_error(self, default: T) -> T:
        """
        Unwraps the value of the monad, if it is a Just.
        Raises the error of the monad, if it is an Error.
        Returns the default value, if the monad is a Nothing.
        """
        if not self:
            return default

        return self._value


with_errorhandling = Maybe.with_errorhandling
Just = Maybe.just
Error = Maybe.error
Nothing = Maybe(None)



from unittest import TestCase

import unittest

from ratisbona_utils.languages import encode_english_short_ordinal


class TestEnglish(TestCase):

    def test_encode_english_short_ordinal_must_provide_ith_for_most_numbers(self):
        for i in range(44, 51, 1200):
            self.assertEqual(encode_english_short_ordinal(i), f"{i}th")

    def test_encode_english_short_ordinal_must_append_st_for_1_21_31(self):
        for i in [1, 21, 31, 41, 51, 61, 71, 81, 91, 101, 121, 131]:
            self.assertEqual(encode_english_short_ordinal(i), f"{i}st")

    def test_encode_short_orginal_must_append_th_for_11_111_211(self):
        for i in [11, 111, 211, 311, 411, 511, 611, 711, 811, 911]:
            self.assertEqual(encode_english_short_ordinal(i), f"{i}th")

    def test_encode_short_orginal_must_append_nd_for_2_22_32_42_(self):
        for i in [2, 22, 32, 42, 52, 62, 72, 82, 92, 102, 122, 132]:
            self.assertEqual(encode_english_short_ordinal(i), f"{i}nd")

    def test_encode_short_orginal_must_append_th_for_12_112_212_312_(self):
        for i in [12, 112, 212, 312, 412, 512, 612, 712, 812, 912, 1012, 1112, 1212, 1312]:
            self.assertEqual(encode_english_short_ordinal(i), f"{i}th")

    def test_encode_short_orginal_must_append_rd_for_3_33_43_53_63_73_83_93_103_123(self):
        for i in [3, 33, 43, 53, 63, 73, 83, 93, 103, 123, 133]:
            self.assertEqual(encode_english_short_ordinal(i), f"{i}rd")

    def test_encode_short_orginal_must_append_th_for_13_113_213_313_413_513_1013_1113(self):
        for i in [13, 113, 213, 313, 413, 513, 1013, 1113, 1213, 1313]:
            self.assertEqual(encode_english_short_ordinal(i), f"{i}th")


if __name__ == '__main__':
    unittest.main()



from dataclasses import dataclass
from enum import Enum

from ratisbona_utils.strings import align, longest_line, Alignment


class LineStyle(Enum):
    SINGLE = 0, [0, 0, 0]
    SINGLE_HEAVY = 1, [1, 0, 0]
    DOUBLE = 2, [2, 2, 0]
    TWO_DASH = 3, [0, 0, 0]
    TWO_DASH_HEAVY = 4, [1, 0, 0]
    THREE_DASH = 5, [0, 0, 0]
    THREE_DASH_HEAVY = 6, [1, 0, 0]
    FOUR_DASH = 7, [0, 0, 0]
    FOUR_DASH_HEAVY = 8, [1, 0, 0]

    def __init__(self, value, replacement_escalations: list[int]):
        super().__init__(value)
        self.intvalue = value
        self.replacement_escalations = replacement_escalations


LINESTYLE_BY_REPLACEMENT = {style.intvalue: style for style in LineStyle}


@dataclass(frozen=True)
class BoxCellDescription:
    top: LineStyle | None
    right: LineStyle | None
    bottom: LineStyle | None
    left: LineStyle | None


def _create_dictionary():
    s = LineStyle.SINGLE
    h = LineStyle.SINGLE_HEAVY
    d2 = LineStyle.TWO_DASH
    d2h = LineStyle.TWO_DASH_HEAVY
    d3 = LineStyle.THREE_DASH
    d3h = LineStyle.THREE_DASH_HEAVY
    d4 = LineStyle.FOUR_DASH
    d4h = LineStyle.FOUR_DASH_HEAVY
    d = LineStyle.DOUBLE
    _ = None

    retval = {}

    def bd(
        top: LineStyle | None,
        right: LineStyle | None,
        bottom: LineStyle | None,
        left: LineStyle | None,
        char: int,
    ):
        retval[BoxCellDescription(top, right, bottom, left)] = char

    bd(_, s, _, s, 0x2500),  # --
    bd(_, h, _, h, 0x2501),
    bd(s, _, s, _, 0x2502),  # |
    bd(h, _, h, _, 0x2503),
    bd(_, d3, _, d3, 0x2504),  # --
    bd(_, d3h, _, d3h, 0x2505),
    bd(d3, _, d3, _, 0x2506),  # |
    bd(d3h, _, d3h, _, 0x2507),

    bd(_, d4, _, d4, 0x2508),  # --
    bd(_, d4h, _, d4h, 0x2509),

    bd(d4, _, d4, _, 0x250A),  # |
    bd(d4h, _, d4h, _, 0x250B),

    bd(_, s, s, _, 0x250C),  # ,-
    bd(_, h, s, _, 0x250D),
    bd(_, s, h, _, 0x250E),
    bd(_, h, h, _, 0x250F),

    bd(_, _, s, s, 0x2510),  # -,
    bd(_, _, h, s, 0x2511),
    bd(_, _, s, h, 0x2512),
    bd(_, _, h, h, 0x2513),

    bd(s, s, _, _, 0x2514),  # '-,
    bd(s, h, _, _, 0x2515),
    bd(h, s, _, _, 0x2516),
    bd(h, h, _, _, 0x2517),

    bd(s, _, _, s, 0x2518),  # -'
    bd(s, _, _, h, 0x2519),
    bd(h, _, _, s, 0x251A),
    bd(h, _, _, h, 0x251B),

    bd(s, s, s, _, 0x251C),  # |-
    bd(s, h, s, _, 0x251D),
    bd(h, s, s, _, 0x251E),
    bd(s, s, h, _, 0x251F),
    bd(h, s, h, _, 0x2520),
    bd(h, h, s, _, 0x2521),
    bd(s, h, h, _, 0x2522),
    bd(h, h, h, _, 0x2523),

    bd(s, _, s, s, 0x2524),  # -|
    bd(s, _, s, h, 0x2525),
    bd(h, _, s, s, 0x2526),
    bd(s, _, s, h, 0x2527),
    bd(h, _, h, s, 0x2528),
    bd(h, _, s, h, 0x2529),
    bd(s, _, h, h, 0x252A),
    bd(h, _, h, h, 0x252B),

    bd(_, s, s, s, 0x252C),  # -,-
    bd(_, s, s, h, 0x252D),
    bd(_, h, s, s, 0x252E),
    bd(_, h, s, h, 0x252F),
    bd(_, s, h, s, 0x2530),
    bd(_, s, h, h, 0x2531),
    bd(_, h, h, s, 0x2532),
    bd(_, h, h, h, 0x2533),

    bd(s, s, _, s, 0x2534),  # -`-
    bd(s, h, _, s, 0x2535),
    bd(h, s, _, s, 0x2536),
    bd(s, h, _, h, 0x2537),
    bd(h, s, _, s, 0x2538),
    bd(h, s, _, h, 0x2539),
    bd(h, h, _, s, 0x253A),
    bd(h, h, _, h, 0x253B),

    bd(s, s, s, s, 0x253C),  #  +
    bd(s, s, s, h, 0x253D),
    bd(s, h, s, s, 0x253E),
    bd(s, h, s, h, 0x253F),
    bd(h, s, s, s, 0x2540),
    bd(s, s, h, s, 0x2541),
    bd(h, s, h, s, 0x2542),
    bd(h, s, s, h, 0x2543),
    bd(h, h, s, s, 0x2544),
    bd(s, s, h, h, 0x2545),
    bd(s, h, h, s, 0x2546),
    bd(h, h, s, h, 0x2547),
    bd(s, h, h, h, 0x2548),
    bd(h, s, h, h, 0x2549),
    bd(h, h, h, s, 0x254A),
    bd(h, h, h, h, 0x254B),

    bd(_, d2, d2, _, 0x254C),  #  -
    bd(_, d2h, d2h, _, 0x254D),
    bd(d2, _, _, d2, 0x254E),
    bd(d2h, _, _, d2h, 0x254F),

    bd(_, d, _, d, 0x2550),  #  ==
    bd(d, _, d, _, 0x2551),  #  ||
    bd(_, d, s, _, 0x2552),  #  ,=
    bd(_, s, d, _, 0x2553),  # ,,-
    bd(_, d, d, _, 0x2554),  #  ,,=
    bd(_, _, s, d, 0x2555),  #  =,
    bd(_, _, d, s, 0x2556),  #  -,,
    bd(_, _, d, d, 0x2557),  #  =,,
    bd(s, d, _, _, 0x2558),  #  '=
    bd(d, s, _, _, 0x2559),  #  ,,-
    bd(d, d, _, _, 0x255A),  #  ,,=
    bd(s, _, _, d, 0x255B),  #  =''
    bd(d, _, _, s, 0x255C),  #  =''
    bd(d, _, _, d, 0x255D),  #  =''
    bd(s, d, s, _, 0x255E),  #  |=
    bd(d, s, d, _, 0x255F),  #  ||-
    bd(d, d, d, _, 0x2560),  #  ||=
    bd(s, _, s, d, 0x2561),  #  =|
    bd(d, _, d, s, 0x2562),  #  -||
    bd(d, _, d, d, 0x2563),  #  =||
    bd(_, s, d, s, 0x2564),  #  -,,-
    bd(_, d, s, d, 0x2565),  #  =,=
    bd(_, d, d, d, 0x2566),  #  =,,=
    bd(s, d, _, d, 0x2567),  #  ='=
    bd(d, s, _, s, 0x2568),  #  -''-
    bd(d, d, _, d, 0x2569),  #  =''=
    bd(s, d, s, d, 0x256A),  #  =|=
    bd(d, s, d, s, 0x256B),  #  -||-
    bd(d, d, d, d, 0x256C),  #  =||=
    return retval


DESCRIPTIONS = _create_dictionary()


def get_char(description: BoxCellDescription) -> str:
    if description in DESCRIPTIONS:
        return chr(DESCRIPTIONS[description])
    for escalation in range(3):
        alternate_description = BoxCellDescription(
            (
                LINESTYLE_BY_REPLACEMENT[
                    description.top.replacement_escalations[escalation]
                ]
                if description.top is not None
                else None
            ),
            (
                LINESTYLE_BY_REPLACEMENT[
                    description.right.replacement_escalations[escalation]
                ]
                if description.right is not None
                else None
            ),
            (
                LINESTYLE_BY_REPLACEMENT[
                    description.bottom.replacement_escalations[escalation]
                ]
                if description.bottom is not None
                else None
            ),
            (
                LINESTYLE_BY_REPLACEMENT[
                    description.left.replacement_escalations[escalation]
                ]
                if description.left is not None
                else None
            ),
        )
        if alternate_description in DESCRIPTIONS:
            return chr(DESCRIPTIONS[alternate_description])
    return "*"


def draw_box(
    multiline_text: str,
    linestyle: LineStyle | str = LineStyle.SINGLE,
    alignment: Alignment = Alignment.LEFT,
    width=None,
) -> str:
    effective_width = width or longest_line(multiline_text) + 4

    if isinstance(linestyle, str):
        vertical = linestyle
        horizontal = linestyle
        top_left = linestyle
        top_right = linestyle
        bottom_left = linestyle
        bottom_right = linestyle
    else:
        vertical = get_char(BoxCellDescription(linestyle, None, linestyle, None))
        horizontal = get_char(BoxCellDescription(None, linestyle, None, linestyle))
        top_left = get_char(BoxCellDescription(None, linestyle, linestyle, None))
        top_right = get_char(BoxCellDescription(None, None, linestyle, linestyle))
        bottom_left = get_char(BoxCellDescription(linestyle, linestyle, None, None))
        bottom_right = get_char(BoxCellDescription(linestyle, None, None, linestyle))

    result = ""
    result += top_left + horizontal * (effective_width - 2) + top_right + "\n"
    for line in multiline_text.splitlines():
        result += vertical
        result += " "
        result += align(line, width=effective_width - 3, alignment=alignment)
        result += vertical
        result += "\n"
    result += bottom_left + horizontal * (effective_width - 2) + bottom_right + "\n"
    return result






import unittest
from datetime import timedelta, date
from multiprocessing import process

import asyncio
from contourpy.util.data import simple
from ratisbona_utils.asyncio import simple_run_command
from ratisbona_utils.datetime import yold_by_date, maybe_get_discordian_season_and_day, maybe_get_discordian_weekday, \
    decode, ddate


class TestDDate(unittest.TestCase):#

    def test_redate_must_be_same_as_translated_date(self):
        for day in range(0, 800):
            a_date = date(2024, 1, 1) + timedelta(days=day)
            yold = yold_by_date(a_date)
            season, day = maybe_get_discordian_season_and_day(a_date)
            weekday = maybe_get_discordian_weekday(a_date)
            redate = decode(day, season, yold)
            self.assertEqual(a_date, redate)

    def test_ddate_must_be_equal_to_cmd_line_tool(self):
        for day in range(0,800):
            # given
            a_date = date(2024, 1, 1) + timedelta(days=day)
            yold = yold_by_date(a_date)
            season, dday = maybe_get_discordian_season_and_day(a_date)
            weekday = maybe_get_discordian_weekday(a_date)

            # when
            string = ddate(dday, season, weekday, yold)


            # then
            # shell invokde ddate
            cmd = f'ddate +"Today is %{{%A, the %e of %B%}} in the YOLD %Y" {a_date.strftime("%d %m %Y")}'
            retval, stdout, stderr = asyncio.run(simple_run_command(cmd))
            print(a_date, season, dday, string, stdout)
            self.assertEqual(stdout.strip(), string)




if __name__ == '__main__':
    unittest.main()



import re
from dataclasses import dataclass
from typing import List, Optional, Generic, TypeVar, Callable, Tuple
import bisect

# Type alias for clarity.
RegularExpression = str
Position = int
LineNum = int
ColNum = int

@dataclass(frozen=True)
class TokenDefinition:
    token_type_name: str
    token_type_regexp: RegularExpression


T = TypeVar("T")


@dataclass(frozen=True)
class Token(Generic[T]):
    token_type_name: str
    token_full_match: str
    token_match_groups: dict[str, str]
    token_match_start_pos: int
    token_match_end_pos: int


def get_tokenizer(token_definitions: List[TokenDefinition]):

    # Precompile the regexes for efficiency.
    compiled_defs = [
        re.compile(tokendef.token_type_regexp) for tokendef in token_definitions
    ]

    def tokenize(text: str) -> List[Token]:

        pos = 0
        while pos < len(text):
            # Track the best match found among all token definitions.
            earliest_match: Optional[re.Match] = None
            best_token_def: Optional[TokenDefinition] = None

            # Check every token definition for a match somewhere in the remaining text.
            for token_def, pattern in zip(token_definitions, compiled_defs):
                match = pattern.search(text, pos)
                if match is None:
                    continue

                # If this is the first match, or if it occurs earlier than the best so far,
                # update the best match.
                if earliest_match is None or match.start() < earliest_match.start():
                    earliest_match = match
                    best_token_def = token_def
                # If two matches start at the same position, pick the one with the longer match.
                elif match.start() == earliest_match.start():
                    if (
                        match.end() - match.start()
                        > earliest_match.end() - earliest_match.start()
                    ):
                        earliest_match = match
                        best_token_def = token_def

            if earliest_match is None:
                # No token definition matched the remainder of the text.
                yield Token(
                    token_type_name="Text",
                    token_full_match=text[pos:],
                    token_match_groups={},
                    token_match_start_pos=pos,
                    token_match_end_pos=len(text),
                )
                break

            # If there is unmatched text before the earliest match, add it as a "Text" token.
            if earliest_match.start() > pos:
                yield Token(
                    token_type_name="Text",
                    token_full_match=text[pos : earliest_match.start()],
                    token_match_groups={},
                    token_match_start_pos=pos,
                    token_match_end_pos=earliest_match.start(),
                )

            # Add the matched token.
            yield Token(
                token_type_name=best_token_def.token_type_name,
                token_full_match=earliest_match.group(0),
                token_match_groups=earliest_match.groupdict(),
                token_match_start_pos=earliest_match.start(),
                token_match_end_pos=earliest_match.end(),
            )
            pos = earliest_match.end()

            # To avoid getting stuck if a pattern matches a zero-length string.
            if pos == earliest_match.start():
                pos += 1

    return tokenize




def create_position_mapper(text: str) -> Callable[[Position], Tuple[LineNum, ColNum]]:
    """
    Create a mapping function from a character index to a (line, column) tuple.
    The mapping function returns 1-indexed line and column numbers.

    Parameters:
      text (str): The text to map positions from.

    Returns:
      Callable[[int], Tuple[int, int]]: A function that maps a character index to (line, column).
    """
    # Precompute newline positions.
    # Adding -1 at the beginning so that the first line starts at position 0.
    newline_positions = [-1] + [i for i, char in enumerate(text) if char == "\n"]

    def map_position(pos: int) -> Tuple[int, int]:
        """
        Map a given character index to a (line, column) tuple.

        Parameters:
          pos (int): The zero-based character index in text.

        Returns:
          Tuple[int, int]: A tuple (line, column) with both numbers starting at 1.
        """
        # Find the line number using bisect.
        # bisect_right finds an insertion point so that newline_positions remains sorted.
        line_index = bisect.bisect_right(newline_positions, pos) - 1
        # The start of the current line is one character after the previous newline.
        line_start = newline_positions[line_index] + 1
        column = pos - line_start
        # Return 1-indexed line and column numbers.
        return line_index + 1, column + 1

    return map_position


# Example usage:
if __name__ == "__main__":
    sample_text = "Hello world!\nThis is a test.\nAnother line."
    pos_mapper = create_position_mapper(sample_text)

    # Let's map a few positions:
    positions = [0, 5, 12, 13, 25, len(sample_text) - 1]
    for pos in positions:
        line, col = pos_mapper(pos)
        print(
            f"Position {pos}: line {line}, column {col}: >{sample_text[max(pos-1,0):pos+2]}<"
        )



import typing
import re


class QuotesParser:

    def __init__(self):
        self.reset_quotes()
        self.reset_linecounter()

    def reset_quotes(self):
        self.quotes_open = False

    def reset_linecounter(self):
        self.line_counter = 0

    def parse(self, lines: typing.Sequence[str]) -> typing.List[str]:
        outputakku = []
        for line in lines:
            outputakku.append(self.parseline(line))
        return outputakku

    def parseline(self, line: str) -> str:

        self.line_counter += 1
        if line.strip() == '' and self.quotes_open:
            print(
                f'Warning: End of paragraph found but quotes still open. Assuming close is missing. Line {self.line_counter}')
            self.quotes_open = False
        workline = line

        # Add guardian-chars in case string starts or ends with a quote.
        workline = 'x' + workline + 'x'

        # We go about replacing existing quotes by splitting the string with the
        # guardian-characters attached at the front and on the back by the existing
        # quotes. As some quotes might be done correctly, we use " '' "` and "'
        # as the quote characters.
        # But as quotes might be quoted by backslashes in tex, we do a negative-
        # lookbehind-match (?<!...)
        # To feature a little more readability, we construct the string from parts.
        theRe = '|'.join([
            '(?<!\\\\)"`',
            '(?<!\\\\)"\'',
            "(?<!\\\\)''",
            '(?<!\\\\)"'
        ])
        parts = re.split(theRe, workline)

        resultline = parts[0]
        for part in parts[1:]:
            resultline += '"\'' if self.quotes_open else '"`'
            resultline += part
            self.quotes_open = not self.quotes_open

        # Remove the guards again
        resultline = resultline[1:-1]

        return resultline



import re
from calendar import monthrange
from datetime import time, date, datetime, timedelta
from typing import Iterator

from dateutil.tz import tzlocal

from ratisbona_utils.monads import Maybe, Just, Nothing


def ensure_timezone(a_datetime: datetime) -> datetime:
    if a_datetime.tzinfo is None:
        return a_datetime.replace(tzinfo=tzlocal())
    return a_datetime

def ensure_no_timezone(a_datetime: datetime) -> datetime:
    return a_datetime.replace(tzinfo=None)

def to_datetime(a_date: date) -> datetime:
    return datetime.combine(a_date, time.min)

def format_ms_HHMMSSss(time_elapsed_millis: int, show_msec: bool = True) -> str:
    """
        Formats a time in milliseconds to a string in the format `HH:MM:SS.mmm`.
    """
    raw_secs, milliseconds = divmod(time_elapsed_millis, 1000)
    raw_min, seconds = divmod(raw_secs, 60)
    hours, minutes = divmod(raw_min, 60)
    retval = f"{hours:02d}:{minutes:02d}:{seconds:02d}"
    if show_msec:
        retval += f".{milliseconds:03d}"
    return retval


def last_day_of_month(given_date: date) -> date:
    """
        Gives the last day of the month of the given date.
        Bitterly missed in the standard library for the actual datatype date, no reason aparent, why they combined
        the int-typed version with unrelated stuff to monthrange...

        Args:
            given_date: date, the date for which the last day of the month is requested.

        Returns:
            date: The last day of the month of the given date.
    """
    last_day = monthrange(given_date.year, given_date.month)[1]
    return date(given_date.year, given_date.month, last_day)


def month_iterator(start: date, end: date) -> Iterator[tuple[int, int]]:
    """
    Provides an iterator that advances month by month from start (inclusive) to end (exclusive) and yields year and month number.

    Args:
        start: date, the day the iterator should start (inclusive).
        end: date, the day the iterator should end (exclusive).

    Returns:
        Tuple[int, int]: Year and month, e.g. (2020, 1) for January 2020.
    """
    current = start.replace(day=1)
    while current < end:
        yield current.year, current.month
        current = (current + timedelta(days=32)).replace(day=1)


def try_parse_leading_isodate(value: str) -> Maybe[tuple[date, str]]:
    """
    Tries to parse a date in the format 'YYYY-MM-DD' from the beginning of the given string.

    Args:
        value: str, the string to parse.

    Returns:
        Maybe[date]: The parsed date or nothing if the string does not start with a valid date.
    """
    match = re.match(r"(\d{4})-(\d{2})-(\d{2})(.*)", value)
    if not match:
        return Nothing
    year, month, day = map(int, match.groups()[0:3])
    rest = match.groups()[3]
    try:
        return Just((date(year, month, day), rest))
    except ValueError:
        return Nothing



from ratisbona_utils.parsing import TokenDefinition

token_definitions = [
    # 1. ISO Datetime (with T or space separator) with optional seconds, milliseconds and timezone.
    # Example: 2024-02-28T14:30, 2024-02-28 14:30:00, 2024-02-28T14:30:00.123, 2024-02-28T14:30:00+02:00
    TokenDefinition(
        "Datetime",
        r"(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})[T ]"
        r"(?P<hour>\d{2}):(?P<minute>\d{2})"
        r"(?:\:(?P<second>\d{2}))?"
        r"(?:\.(?P<millisecond>\d{1,3}))?"
        r"(?P<timezone>Z|[+\-]\d{2}:\d{2})?",
    ),
    # 2. German Date: day.month.year (e.g., 28.2.2024)
    TokenDefinition("Date", r"(?P<day>\d{1,2})\.(?P<month>\d{1,2})\.(?P<year>\d{4})"),
    # 3. ISO Date: year-month-day (e.g., 2024-02-28)
    TokenDefinition("Date", r"(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})"),
    # 4. American Date: month/day/year (e.g., 02/28/2024)
    TokenDefinition("Date", r"(?P<month>\d{1,2})/(?P<day>\d{1,2})/(?P<year>\d{4})"),
    # 5. Time (only): hour:minute with optional seconds and milliseconds.
    # Examples: 14:30, 14:30:00, 14:30:00.123
    TokenDefinition(
        "Time",
        r"(?P<hour>\d{1,2}):(?P<minute>\d{2})"
        r"(?:\:(?P<second>\d{2}))?"
        r"(?:\.(?P<millisecond>\d{1,3}))?",
    ),
    # 6. Relative Dates in German and English.
    # Matches words like "heute", "morgen", "übermorgen", "gestern", "vorgestern" as well as
    # "today", "tomorrow", "day after tomorrow", "day before yesterday".
    # Using the inline (?i) flag to ignore case.
    TokenDefinition(
        "RelativeDate",
        r"(?i)\b(?P<rel>(heute|morgen|übermorgen|gestern|vorgestern|today|tomorrow|day after tomorrow|day before yesterday))\b",
    ),
]



from typing import TypeVar

from ratisbona_utils.markdown import Heading
from ratisbona_utils.monads.monads import ResultMonad
from ratisbona_utils.parsing import TokenStream, LineToken, ParseError

T = TypeVar("T")


def _err(
    message: str, line_number: int, near_pos: int | None
) -> ResultMonad[T, ParseError]:
    return ResultMonad[T, ParseError].err(
        ParseError(
            message=message,
            line_number=line_number,
            near_pos=near_pos,
        )
    )


def try_parse_heading(
    line_token_stream: TokenStream[LineToken],
) -> ResultMonad[Heading, ParseError]:
    """
    Tries to parse a heading from the given token stream.

    Args:
        line_token_stream: The token stream to parse from.

    Returns:
        ResultMonad[Heading, ParseError]: A result monad containing the parsed heading or a parse error.
    """
    line_token = line_token_stream.next_token()
    if not line_token.text.startswith("#"):
        line_token_stream.put_back(line_token)
        return _err(
            f"Expected a heading to begin with # but got {line_token[:10]}",
            line_token.line_number, 0
        )

    level = 1
    while line_token.text[level] == "#":
        level += 1

    text = line_token.text[level:].strip()
    return ResultMonad.ok(Heading(level=level, text=text))

def try_parse_paragraph(
    line_token_stream: TokenStream[LineToken],
) -> ResultMonad[Paragraph, ParseError]:
    """
    Tries to parse a paragraph from the given token stream.

    Args:
        line_token_stream: The token stream to parse from.

    Returns:
        ResultMonad[Paragraph, ParseError]: A result monad containing the parsed paragraph or a parse error.
    """
    line_token = line_token_stream.next_token()
    if line_token.text.startswith("#"):
        line_token_stream.put_back(line_token)
        return _err(
            f"Expected a paragraph but got a heading",
            line_token.line_number, 0
        )

    content = []
    while line_token.text:
        content.append(PureText(line_token.text))
        line_token = line_token_stream.next_token()

    return ResultMonad.ok(Paragraph(content=content))


import unittest
from time import sleep

from ratisbona_utils.datetime import Stopwatch, with_stopwatch


class TestStopWatch(unittest.TestCase):

    def test_spending_time_must_advance_clock(self):

        # Given
        stopwatch = Stopwatch()
        watched_sleep = with_stopwatch(sleep, stopwatch)

        # When
        watched_sleep(0.1)

        # Then
        self.assertGreater(stopwatch.time_elapsed_millis, 90)
        print(stopwatch)

    def test_stopwatch_must_not_run_after_watched_function_terminates(self):
        # Given
        stopwatch = Stopwatch()
        watched_sleep = with_stopwatch(sleep, stopwatch)

        # When
        watched_sleep(0.1)
        sleep(1.0)

        # Then
        self.assertGreater(stopwatch.time_elapsed_millis, 90)
        self.assertLess(stopwatch.time_elapsed_millis, 110)

    def test_stopwatch_must_resume_on_multiple_invocations(self):
        # Given
        stopwatch = Stopwatch()
        watched_sleep = with_stopwatch(sleep, stopwatch)

        # When
        watched_sleep(0.1)
        sleep(0.5)
        watched_sleep(0.1)

        # Then
        self.assertGreater(stopwatch.time_elapsed_millis, 180)
        self.assertLess(stopwatch.time_elapsed_millis, 230)
        print(stopwatch)

    def test_suspending_the_clock_must_not_advance_time(self):
        with Stopwatch() as stopwatch:
            sleep(0.1)
            stopwatch.stop()
            sleep(1.0)
            stopwatch.start()
            sleep(0.1)
        self.assertGreater(stopwatch.time_elapsed_millis, 180)
        self.assertLess(stopwatch.time_elapsed_millis, 220)


if __name__ == '__main__':
    unittest.main()



from collections import defaultdict
from datetime import date, timedelta
import matplotlib.pyplot as plt


def calculate_easter_date(year_x: int) -> date:

    saekularzahl_k = year_x // 100
    mondparameter_a = year_x % 19

    zweimal_benoetigt = (3 * saekularzahl_k + 3) // 4

    saekulare_mondschaltung_m = 15 + zweimal_benoetigt - (8 * saekularzahl_k + 13) // 25
    saekulare_sonnenschaltung_s = 2 - zweimal_benoetigt

    keim_des_ersten_vollmonds_im_fruehling_d = (
        19 * mondparameter_a + saekulare_mondschaltung_m
    ) % 30
    kalendarische_korrekturgroesse_r = (
        keim_des_ersten_vollmonds_im_fruehling_d + mondparameter_a // 11
    ) // 29
    erster_sonntag_im_maerz_sz = 7 - (year_x + year_x // 4 + saekulare_sonnenschaltung_s) % 7

    ostergrenze_og = 21 + keim_des_ersten_vollmonds_im_fruehling_d - kalendarische_korrekturgroesse_r

    osterentfernung_von_ostergrenze_oe = 7 - (ostergrenze_og - erster_sonntag_im_maerz_sz) % 7

    ostersonntag_als_maerzdatum = ostergrenze_og + osterentfernung_von_ostergrenze_oe

    return date(year_x, 3, 1) + timedelta(days=ostersonntag_als_maerzdatum - 1)





import unittest
from multiprocessing.pool import ExceptionWithTraceback

from ratisbona_utils.monads import Maybe, Just, Nothing, Error, UnexpectedNothingError


class TestMaybe(unittest.TestCase):

    def test_just_must_be_just(self):
        self.assertTrue(Just(10).is_just)

    def test_just_must_not_be_nothing(self):
        self.assertFalse(Just(10).is_nothing)

    def test_just_must_not_be_error(self):
        self.assertFalse(Just(10).is_error)

    def test_just_must_unwrap_value(self):
        self.assertEqual(Just(10).unwrap_value(), 10)

    def test_just_must_raise_value_error_for_none(self):
        with self.assertRaises(ValueError):
            Just(None)

    def test_just_must_raise_value_error_for_error(self):
        with self.assertRaises(ValueError):
            Just(ValueError("An error occurred"))

    def test_just_must_not_unwrap_error(self):
        with self.assertRaises(ValueError):
            Just(10).unwrap_error()

    def test_just_must_be_true(self):
        self.assertTrue(Just(10))

    def test_just_or_error_must_be_just(self):
        self.assertTrue(Just(10) | Error(ValueError("An error occurred")).is_just)

    def test_just_or_just_must_be_same_just(self):
        self.assertEqual(10, (Just(10) | Just(20)).unwrap_value())

    def test_just_or_nothing_must_be_same_just(self):
        self.assertEqual(10, (Just(10) | Nothing).unwrap_value())

    def test_just_of_just_must_just_be_just(self):
        self.assertTrue(10, Just(Just(10)).unwrap_value())

    def test_just_must_be_same_as_value(self):
        self.assertEqual(Just(10), 10)

    def test_just_must_be_same_as_other_just_if_value_is_equal(self):
        self.assertEqual(Just(10), Just(5 + 5))

    def test_just_must_be_same_as_maybe_value(self):
        self.assertEqual(Just(10), Maybe(10))

    def test_error_must_be_error(self):
        self.assertTrue(Error(ValueError("An error occurred")).is_error)

    def test_error_must_not_be_just(self):
        self.assertFalse(Error(ValueError("An error occurred")).is_just)

    def test_error_must_not_be_nothing(self):
        self.assertFalse(Error(ValueError("An error occurred")).is_nothing)

    def test_error_must_unwrap_error(self):
        self.assertEqual(
            "An test error occurred",
            Error(ValueError("An test error occurred")).unwrap_error().args[0],
        )

    def test_error_must_raise_value_error_for_just_value(self):
        with self.assertRaises(ValueError) as ve:
            Error(42)

    def test_error_must_raise_value_error_for_none(self):
        with self.assertRaises(ValueError) as ve:
            Error(None)

    def test_error_must_not_unwrap_value(self):
        with self.assertRaises(ValueError):
            Error(ValueError("An error occurred")).unwrap_value()

    def test_error_must_not_be_true(self):
        self.assertFalse(Error(ValueError("An error occurred")))

    def test_error_or_error_must_be_second_error(self):
        self.assertEqual(
            "Another error occurred",
            Error(ValueError("An error occurred"))
            | Error(ValueError("Another error occurred")).unwrap_error().args[0],
        )

    def test_error_or_just_must_be_just(self):
        self.assertTrue((Error(ValueError("An error occurred")) | Just(10)).is_just)

    def test_error_or_nothing_must_be_nothing(self):
        self.assertTrue((Error(ValueError("An error occurred")) | Nothing).is_nothing)

    def test_error_of_error_must_be_error(self):
        self.assertTrue(
            "An test error occurred",
            Error(Error(Error(ValueError("An test error occurred"))))
            .unwrap_error()
            .args[0],
        )

    def test_error_must_be_same_as_value(self):
        error = ValueError("A test error occurred")
        self.assertEqual(Error(error), error)

    def test_error_must_be_same_as_other_error_if_exceptions_are_equal(self):
        error = ValueError("A test error occurred")
        self.assertEqual(Error(error), Error(error))

    def test_error_must_be_same_as_maybe_error(self):
        error = ValueError("A test error occurred")
        self.assertEqual(Error(error), Maybe(error))

    def test_nothing_must_be_nothing(self):
        self.assertTrue(Nothing.is_nothing)

    def test_nothing_must_not_be_just(self):
        self.assertFalse(Nothing.is_just)

    def test_nothing_must_not_be_error(self):
        self.assertFalse(Nothing.is_error)

    def test_nothing_must_not_unwrap_value(self):
        with self.assertRaises(ValueError):
            Nothing.unwrap_value()

    def test_nothing_must_not_unwrap_error(self):
        with self.assertRaises(ValueError):
            Nothing.unwrap_error()

    def test_nothing_must_not_be_true(self):
        self.assertFalse(Nothing)

    def test_nothing_or_error_must_be_error(self):
        self.assertTrue((Nothing | Error(ValueError("An error occurred"))).is_error)

    def test_nothing_or_just_must_be_just(self):
        self.assertTrue((Nothing | Just(10)).is_just)

    def test_nothing_or_nothing_must_be_nothing(self):
        self.assertTrue((Nothing | Nothing).is_nothing)

    def test_nothing_must_equal_None(self):
        self.assertEqual(Nothing, None)

    def test_nothing_must_equal_other_nothing(self):
        self.assertEqual(Nothing, Nothing)

    def test_nothing_must_equal_maybe_none(self):
        self.assertEqual(Nothing, Maybe(None))

    def test_bind_to_just_must_apply_function(self):
        self.assertEqual(Just(10).bind(lambda x: x + 5).unwrap_value(), 15)

    def test_bind_to_nothing_must_return_nothing(self):
        self.assertTrue(Nothing.bind(lambda x: x + 5).is_nothing)

    def test_bind_to_error_must_return_error(self):
        self.assertTrue(
            Error(ValueError("An error occurred")).bind(lambda x: x + 5).is_error
        )

    def test_bind_to_just_of_function_raising_error_must_return_error(self):
        maybe_result = Just(10).bind(lambda x: x / 0)
        self.assertTrue(maybe_result.is_error)
        the_error = maybe_result.unwrap_error()
        self.assertIsInstance(the_error, ZeroDivisionError)
        self.assertTrue("Traceback" in the_error.__notes__[0])


    def test_binding_just_function_to_just_must_apply_function(self):
        def divide_by_2(x):
            return x / 2

        maybe_divide_by_2 = Just(divide_by_2)

        self.assertEqual(Just(10).bind(maybe_divide_by_2).unwrap_value(), 5)

    def test_binding_nothing_function_to_just_must_return_nothing(self):
        self.assertTrue(Just(10).bind(Nothing).is_nothing)

    def test_binding_error_function_to_just_must_return_error(self):
        error_function = Error(ValueError("An error occurred"))
        self.assertTrue(Just(10).bind(error_function).is_error)

    def test_binding_must_accept_additional_arguments(self):
        def add(x, y):
            return x + y

        self.assertEqual(Just(10).bind(add, 5).unwrap_value(), 15)

    def test_binding_additional_maybe_must_unwrap(self):
        def add(x, y):
            return x + y

        self.assertEqual(Just(10).bind(add, Just(5)).unwrap_value(), 15)

    def test_binding_additional_nothing_must_return_nothing(self):
        def add(x, y):
            return x + y

        self.assertTrue(Just(10).bind(add, Nothing).is_nothing)

    def test_binding_additional_error_must_return_error(self):
        def add(x, y):
            return x + y

        self.assertTrue(
            Just(10).bind(add, Error(ValueError("An error occurred"))).is_error
        )

    def test_in_binding_errors_of_original_monad_must_be_dominant(self):
        originally_in_error = Error(ValueError("The original error"))
        originally_nothing = Nothing

        erratic_function = Error(ValueError("Erratic function error"))
        erratic_argument = Error(ValueError("Erratic argument error"))

        self.assertEqual(
            "The original error",
            originally_in_error.bind(erratic_function).unwrap_error().args[0],
        )
        self.assertEqual(
            "The original error",
            originally_in_error.bind(lambda x: x, erratic_argument)
            .unwrap_error()
            .args[0],
        )
        self.assertTrue(originally_nothing.bind(erratic_function).is_nothing)
        self.assertTrue(
            originally_nothing.bind(lambda x: x, erratic_argument).is_nothing
        )

    def test_maybe_recover_must_recover_on_just_returnvalue(self):
        error_value = Error(ValueError("An error occurred"))
        result = error_value.maybe_recover(lambda e: 20)
        self.assertTrue(result.is_just)
        self.assertFalse(result.is_error)
        self.assertFalse(result.is_nothing)
        self.assertEqual(result.unwrap_value(), 20)

    def test_maybe_recover_must_not_recover_on_error_returnvalue(self):
        error_value = Error(ValueError("An error occurred"))
        result = error_value.maybe_recover(lambda e: Error(ValueError("Another error")))
        self.assertTrue(result.is_error)
        self.assertFalse(result.is_just)
        self.assertFalse(result.is_nothing)
        self.assertEqual(result.unwrap_error().args[0], "An error occurred")

    def test_maybe_recover_must_recover_on_nothing_returnvalue(self):
        error_value = Error(ValueError("An error occurred"))
        result = error_value.maybe_recover(lambda e: Nothing)
        self.assertTrue(result.is_nothing)

    def test_default_or_throw_must_return_value(self):
        self.assertEqual(10, Just(10).default_or_throw(20))

    def test_default_or_throw_must_raise_error(self):
        with self.assertRaises(ValueError):
            Error(ValueError("An error occurred")).default_or_throw(20)

    def test_default_or_throw_must_substitute_default_on_nothing(self):
        self.assertEqual(20, Nothing.default_or_throw(20))

    def test_default_on_error_must_return_value(self):
        self.assertEqual(10, Just(10).default_also_on_error(20))

    def test_default_on_error_must_return_default_on_error(self):
        self.assertEqual(20, Error(ValueError("An error occurred")).default_also_on_error(20))

    def test_default_on_error_must_return_default_on_nothing(self):
        self.assertEqual(20, Nothing.default_also_on_error(20))

    def test_value_or_throw_must_return_value(self):
        self.assertEqual(10, Just(10).value_or_throw())

    def test_value_or_throw_must_raise_error_on_error(self):
        with self.assertRaises(ValueError):
            Error(ValueError("An error occurred")).value_or_throw()

    def test_value_or_throw_must_raise_error_on_nothing(self):
        with self.assertRaises(UnexpectedNothingError) as une:
            Nothing.value_or_throw()
        note = une.exception.__notes__[0]
        print(note)
        self.assertTrue("Traceback" in note)

    def test_getitem_must_return_value(self):
        maybe_dict = Just({"a": 10, "b": 20})
        maybe_a = maybe_dict.__getitem__("a")
        a_value = maybe_a.unwrap_value()
        self.assertEqual(10, a_value)

        self.assertEqual(10, maybe_dict["a"].unwrap_value())

    def test_iterating_over_maybe_of_iterable_must_result_in_maybe_items(self):
        maybe_list = Just([1, 2, 3])
        for item, expected in zip(maybe_list, [1,2,3]):
            self.assertIsInstance(item, Maybe)
            self.assertTrue(item.is_just)
            self.assertEqual(expected, item.unwrap_value())

    def test_iterating_over_nothing_must_not_iterate(self):
        for _ in Nothing:
            self.fail("Nothing should not iterate")

    def test_iterating_over_error_must_not_iterate(self):
        for _ in Error(ValueError("An error")):
            self.fail("Error should not iterate")

    def test_iterating_over_single_element_should_iterate_exactly_once(self):
        for item in Just(10):
            self.assertEqual(10, item.unwrap_value())

    def test_in_must_return_true_if_item_in_monads_value(self):
        self.assertTrue(10 in Just([9,10,11]))
        self.assertTrue("a" in Just({"a": 10, "b": 20}))

    def test_in_must_return_false_if_item_not_in_monads_value(self):
        self.assertFalse(10 in Just([9,11]))
        self.assertFalse("a" in Just({"b": 10, "c": 20}))

    def test_in_must_return_true_if_value_of_item_in_monads_value(self):
        self.assertTrue(Just(10) in Just([9,10,11]))
        self.assertTrue(Just("a") in Just({"a": 10, "b": 20}))

    def test_in_must_be_false_for_nothing(self):
        self.assertFalse(10 in Nothing)
        self.assertFalse("a" in Nothing)

    def test_in_must_be_false_for_error(self):
        self.assertFalse(10 in Error(ValueError("An error")))
        self.assertFalse("a" in Error(ValueError("An error")))

    def test_in_must_be_false_for_nothing_item(self):
        self.assertFalse(Nothing in Just([9,10,11]))
        self.assertFalse(Nothing in Just({"a": 10, "b": 20}))

    def test_in_must_be_false_for_error_item(self):
        self.assertFalse(Error(ValueError("An error")) in Just([9,10,11]))
        self.assertFalse(Error(ValueError("An error")) in Just({"a": 10, "b": 20}))









if __name__ == "__main__":
    unittest.main()



import sys
from pathlib import Path

UTF8 = {"encoding": "utf-8"}

def errprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

def maybe_backup_file(filepath: Path) -> bool:
    if not filepath.exists():
        return False
    print(f"Backuping {filepath}")
    with (
        filepath.with_suffix(".bak").open("wb") as backup,
        filepath.open("rb") as file
    ):
        backup.write(file.read())
    return True


import inspect
from datetime import datetime
from pathlib import Path

import click

import ratisbona_utils.colors.good_palettes as good_palettes
from ratisbona_utils.colors import (
    parse_gimp_palette,
    to_gimp_palette,
    palette_generators,
    Palette,
    rgb_to_hex,
)
from ratisbona_utils.colors.serde_json_palette import to_json, from_json
from ratisbona_utils.terminals.vt100 import color_block


def do_display_palette(palette) -> None:
    for color, name, colortype in zip(
        palette.colors, palette.color_names, palette.color_types
    ):
        displaycolor = rgb_to_hex(color) if colortype == "rgb" else str(color)
        block = color_block(color)
        print(f"{block}", end="")

def display_named_colors(palette: Palette) -> None:
    for color, name, colortype in zip(
        palette.colors, palette.color_names, palette.color_types
    ):
        displaycolor = rgb_to_hex(color) if colortype == "rgb" else str(color)
        block = color_block(color)
        print(f"{block} {name} {displaycolor}")

@click.group()
def palette_cli():
    print("*" * 80)
    print("   Palette CLI")
    print("*" * 80)


@click.command("parse-gimp")
@click.argument(
    "gimp_palette_file",
    type=click.Path(exists=True, path_type=Path, dir_okay=False, file_okay=True),
)
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=False, path_type=Path, dir_okay=False, file_okay=True),
    required=False,
    default=None,
)
def gimp_palette_parser(gimp_palette_file: Path, output: Path | None):
    """
    Parse a GIMP palette file and print the palette object.
    """
    palette_str = gimp_palette_file.read_text()
    gimp_palette = parse_gimp_palette(palette_str)
    json_text = to_json(gimp_palette)
    if output:
        output.write_text(json_text)
    else:
        print(json_text)


palette_cli.add_command(gimp_palette_parser)


@click.command("write-gimp")
@click.argument(
    "native_palette",
    type=click.Path(exists=True, path_type=Path, dir_okay=False, file_okay=True),
)
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=False, path_type=Path, dir_okay=False, file_okay=True),
    required=False,
    default=None,
)
def gimp_palette_writer(native_palette: Path, output: Path | None):
    """
    Write a GIMP palette file from a palette object.
    """
    palette_json = native_palette.read_text()
    palette = from_json(palette_json)
    gimp_palette = to_gimp_palette(palette)
    if output:
        output.write_text(gimp_palette)
    else:
        print(gimp_palette)


palette_cli.add_command(gimp_palette_writer)


PALETTES = [
    identifier for identifier in dir(good_palettes) if identifier.endswith("_PALETTE")
]


@click.command("writeout-builtin")
@click.argument("palette_name", type=click.Choice(PALETTES))
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=False, path_type=Path, dir_okay=False, file_okay=True),
    required=False,
    default=None,
)
def writeout_builtin_palette(palette_name: str, output: Path | None):
    """
    Write a builtin palette to a file.
    """
    palette = getattr(good_palettes, palette_name)
    json_str = to_json(palette)
    do_display_palette(palette)
    if output:
        output.write_text(json_str)
    else:
        print(json_str)


palette_cli.add_command(writeout_builtin_palette)

GENERATORS = ["viridis", "cielch_based_rgb_palette", "hsv_based_rgb_palette", "terrain"]


@click.command("writeout-generated")
@click.argument("generator", type=click.Choice(GENERATORS))
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=False, path_type=Path, dir_okay=False, file_okay=True),
    required=False,
    default=None,
)
@click.option(
    "--hue-steps",
    "-n",
    type=int,
    required=False,
    default=16,
)
@click.option(
    "--lightness",
    "-l",
    type=float,
    required=False,
    default=50,
)
@click.option(
    "--chroma",
    "-c",
    type=float,
    required=False,
    default=50,
)
@click.option(
    "--saturation",
    "-s",
    type=float,
    required=False,
    default=1,
)
@click.option(
    "--value",
    "-v",
    type=float,
    required=False,
    default=1,
)
def writeout_generated_palette(
    generator, output: Path | None, hue_steps, lightness, chroma, saturation, value
):
    """
    Write a generated palette to a file.

    Args:
        generator:
        output:
        hue_steps:
        lightness:
        chroma:
        saturation:
        value:

    Returns:
    """
    provided_args = {
        "hue_steps": hue_steps,
        "lightness": lightness,
        "chroma": chroma,
        "saturation": saturation,
        "value": value,
    }
    palette = getattr(palette_generators, generator)
    signature = inspect.signature(palette)
    valid_args = {
        key: value
        for key, value in provided_args.items()
        if key in signature.parameters
    }
    bound_args = signature.bind_partial(**valid_args)
    rgb_colors = palette(**bound_args.arguments)
    palette = Palette(
        name=f"{generator}_palette",
        description=f"Generated with {generator}",
        author="palette_cli",
        creation_date=datetime.now(),
        color_names=[""] * len(rgb_colors),
        color_types=["rgb"] * len(rgb_colors),
        colors=rgb_colors,
    )
    json_str = to_json(palette)
    do_display_palette(palette)
    if output:
        output.write_text(json_str)
    else:
        print(json_str)


palette_cli.add_command(writeout_generated_palette)


@click.command("display")
@click.argument(
    "palette_file",
    type=click.Path(exists=True, path_type=Path, dir_okay=False, file_okay=True),
)
def display_palette(palette_file: Path):
    """
    Display a palette file.
    """
    palette = from_json(palette_file.read_text())
    display_named_colors(palette)
    do_display_palette(palette)



palette_cli.add_command(display_palette)



from pathlib import Path
from unittest import TestCase

from ratisbona_utils.images import ImageCache
from shutil import rmtree

url='https://www.duh.de/fileadmin/_processed_/f/a/csm_woods_ad5acec37a.png'

class TestImageCache(TestCase):

    def test_get_file_for_url_must_yield_path(self):
        path=Path('/tmp/imgcachetest')
        if path.exists():
            #recursively delete directory with content
            rmtree(path)
        with ImageCache(path) as mgr:
            sudopath=mgr.get_file_for_url(url)
        self.assertEqual(sudopath, path / '0')

    def test_downloading_must_fetch_files(self):
        path=Path('/tmp/imgcachetest')
        if path.exists():
            #recursively delete directory with content
            rmtree(path)
        with ImageCache(path) as mgr:
            sudopath=mgr.get_file_for_url(url)
            mgr.download_all()
        self.assertTrue(sudopath.exists())


from unittest import TestCase

import asyncio
from ratisbona_utils.asyncio._process_runner import simple_run_command


class TestProcessRunner(TestCase):

    def test_simple_run_process(self):
        cmd = 'echo "Hello, World!"; echo "Hollas!" >&2'

        result, stdout, stderr = asyncio.run(simple_run_command(cmd))

        self.assertEqual(result, 0)
        self.assertEqual(stdout, 'Hello, World!\n')
        self.assertEqual(stderr, 'Hollas!\n')



import json
from datetime import datetime

from ratisbona_utils.colors import Palette, rgb_to_hex, hex_to_rgb
from ratisbona_utils.json import (
    simple_json_serializer,
    to_json_line,
    to_json_array_horizontal, to_json_object,
)


def to_json(palette: Palette) -> str:
    palette_dict = palette.__dict__
    ordinary_lines = [
        to_json_line(key, json.dumps(value))
        for key, value in palette_dict.items()
        if key not in ["colors", "color_names", "color_types", "creation_date"]
    ]
    creation_date_line = to_json_line(
        "creation_date",
        json.dumps(palette.creation_date.isoformat() if palette.creation_date else None),
    )
    color_names_line = to_json_line(
        "color_names",
        to_json_array_horizontal([json.dumps(name) for name in palette.color_names]),
    )
    color_types_line = to_json_line(
        "color_types",
        to_json_array_horizontal([json.dumps(type_) for type_ in palette.color_types]),
    )

    mangled_colors = []
    for color, colortype in zip(palette.colors, palette.color_types):
        if colortype == "rgb":
            mangled_colors.append(json.dumps(rgb_to_hex(color)))
        else:
            mangled_colors.append(json.dumps(color))
    colors_line = to_json_line("colors", to_json_array_horizontal(mangled_colors))
    return to_json_object(ordinary_lines + [creation_date_line, color_names_line, color_types_line, colors_line])


def from_json(json_text: str) -> Palette:
    """
    Deserialize a Palette object from a JSON string.
    Args:
        json_text: The JSON string to deserialize.

    Returns:
        Palette: The deserialized Palette object.
    """
    attributes = json.loads(json_text)

    if attributes["creation_date"]:
        attributes["creation_date"] = datetime.fromisoformat(attributes["creation_date"])

    remangled_colors = []
    for color, colortype in zip(attributes["colors"], attributes["color_types"]):
        if colortype == "rgb":
            remangled_colors.append(hex_to_rgb(color))
        else:
            remangled_colors.append(color)
    attributes["colors"] = remangled_colors
    return Palette(**attributes)





# coding: utf-8
import os
import urllib.request
import json
from pathlib import Path


class ImageCache:
    """
        This class is a simple image cache that stores images locally.
        It is intended to be used as a context manager. So use it like this:

            ``` python
            with ImageCache(Path('/tmp/imgcachetest')) as mgr:
                local_file=mgr.get_file_for_url(url)
                mgr.download_all()
            ```

        The `local_file` will contain the local path to the image.

    """
    
    def __init__(self, dirname: Path):
        """
            Constructor.
        Args:
            dirname: The directory where the images are stored.
        """
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        self.jsoncontent = {}
        self.dirname = dirname
        self.indexfile = self.get_local_file_by_count('index.json')
        if os.path.exists(self.indexfile):
            with open(self.indexfile, 'r') as file:
                self.jsoncontent = json.loads(file.read())
        
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        with open(self.indexfile, 'w') as file:
            file.write(json.dumps(self.jsoncontent, indent=2))
    
    def add_url(self, url) -> Path:
        """
            Adds a URL to the cache.
        Args:
            url: The URL to add.

        Returns:
            The local path to the image.
        """
        countval = self.jsoncontent.get('counter',-1)+1
        self.jsoncontent.update({'counter': countval})
        name_map = self.jsoncontent.get('nameMap',{})
        name_map.update({url: countval})
        self.jsoncontent.update({'nameMap': name_map})
        return self.get_local_file_by_count(str(countval))
        
    def get_file_for_url(self, url) -> Path:
        """
            Gets the local file for a URL, if downloading was successful.
            If download was not successful, the file will not exist.
            It might turn up later, if you call `download_all` again, because it will retry failed downloads.
        Args:
            url: The URL to get the local file for.

        Returns:
            The local path to the image.
        """
        nameMap = self.jsoncontent.get('nameMap',{})
        if url in nameMap:
            return self.get_local_file_by_count(nameMap.get(url))
        return self.add_url(url)
    
    def get_local_file_by_count(self, count) -> Path:
        return self.dirname / str(count)
    
    def download_all(self, retry_failed=False):
        """
            Downloads all images that are not yet downloaded.
        Args:
            retry_failed: False by default. If true, failed downloads will be retried.

        Returns:
            Nothing
        """
        name_map = self.jsoncontent.get('nameMap', {})
        download_map = self.jsoncontent.get('downloadMap',{})
        total = len(name_map)
        count = 0
        for url,countval in name_map.items():
            if url is None: continue
            map_entry = download_map.get(url, None)
            print(
                'Considering {} Mapentry: {} ({}/{})'.format(
                    url, map_entry, count, total
                )
            )
            if map_entry is None or map_entry == 'FAILED' and retry_failed:
                filename = self.get_local_file_by_count(str(countval))
                if os.path.exists(filename):
                    download_map.update({url: 'OK'})
                else:
                    try:
                        print("Downloading: " + url)
                        urllib.request.urlretrieve(url, filename)
                        download_map.update({url: 'OK'})
                    except Exception as e:
                        download_map.update({url: 'FAILED'})
            self.jsoncontent.update({'downloadMap': download_map})
            count += 1
        

      
        
        



import datetime as dt
from itertools import islice
from random import Random


def autoseed_16bit() -> int:
    return int(
        dt.datetime.now().microsecond
        * dt.datetime.now().timestamp()
    ) & 0xFFFF_FFFF


def linearCongruentialGenerator(
    modulus_mask: int, coefficient: int, constant: int, seed: int
):
    while True:
        seed = (seed * coefficient + constant) & modulus_mask
        yield seed & modulus_mask


def knuth_mmix(seed):
    return linearCongruentialGenerator(
        0xFFFFFFFF_FFFFFFFF,
        6364136223846793005,
        1442695040888963407,
        seed,
    )

def normalized_generator(generator, modulus_mask: int) -> float:
    while True:
        try:
            yield next(generator) / modulus_mask
        except StopIteration:
            return

def between(lower, upper, normalized_generator):
    while True:
        try:
            yield next(normalized_generator) * (upper - lower) + lower
        except StopIteration:
            return


def lsfr_generator(
    seed: int, polynomal_mask: int
):
    while True:
        seed = (seed >> 1) ^ (-(seed & 1) & polynomal_mask)
        yield seed

def lsfr_B4BCD35C_32Bit(seed: int):
    return lsfr_generator(seed, 0xB4BCD35C)


def int_between(lower, upper, normalized_generator):
    while True:
        try:
            yield int(next(normalized_generator) * (upper - lower + 1) + lower)
        except StopIteration:
            return




from ratisbona_utils.colors.simple_color import RGBColor
from ratisbona_utils.colors.palette import Palette


def to_gimp_palette(palette: Palette[RGBColor]) -> str:
    """
    Convert a Palette object to a GIMP palette file format.

    Args:
        palette (Palette): The palette object to convert

    Returns:
        str: The GIMP palette file content
    """
    description = palette.description.replace('\n', ' ')
    lines = [
        "GIMP Palette",
        f"Name: {palette.name}",
        "Columns: 16",
        f"# Description: {description}",
        f"# Author: {palette.author}",
        f"# Created: {palette.creation_date}",
    ]
    for color, name in zip(palette.colors, palette.color_names):
        lines.append(f"{color[0]} {color[1]} {color[2]} {name}")
    return "\n".join(lines)


def parse_gimp_palette(palette_as_str: str) -> Palette:
    """
    Parse a GIMP palette file and return the Palette object.

    Args:
        palette_as_str (str): The GIMP palette file content as a string

    Returns:
        Palette: The parsed palette object
    """
    lines = palette_as_str.splitlines()
    lines = map(str.strip, lines)
    lines = list(filter(lambda s: s and not s.startswith("#"), lines))
    if not lines[0] == "GIMP Palette":
        raise ValueError("Not a GIMP palette file")
    if not lines[1].startswith("Name: "):
        raise ValueError("Invalid palette name line {line[1]}. Does not start with 'Name: '")
    name = lines[1][6:]
    money_shot_at = 3
    if not lines[2].startswith("Columns: "):
        money_shot_at = 2
        print("Warning: Invalid columns line {line[2]}. Does not start with 'Columns: '")


    colors = []
    names = []
    for line in lines[money_shot_at:]:
        if not line:
            continue
        color = line.split()
        colors.append((int(color[0]), int(color[1]), int(color[2])))
        names.append(color[3] if len(color) > 3 else "")
    return Palette(
        name=name,
        description="",
        author="",
        creation_date=None,
        colors=colors,
        color_names=names,
        color_types=["rgb"] * len(colors)
    )





from unittest import TestCase
from ratisbona_utils.trees import TreePath, TreePathElem, TreeBuilder, GraphVizitor


class TreeTest(TestCase):
    def test_creating_treepath_must_result_in_empty_path(self):
        path = TreePath()
        self.assertEqual(str(path), "/")
        self.assertEqual(path.path_elems, [])

    def test_creating_treepath_by_parsing_single_element_must_result_in_root_with_one_subelement(
        self,
    ):
        path = TreePath.parse("home")
        self.assertEqual(str(path), "/home")
        self.assertEqual(path.path_elems, [TreePathElem("home")])

    def test_parsing_path_with_multiple_elements_must_result_in_correct_path(self):
        path = TreePath.parse("home/kevin")
        self.assertEqual(str(path), "/home/kevin")
        self.assertEqual(path.path_elems, [TreePathElem("home"), TreePathElem("kevin")])

    def test_creating_path_from_string_must_result_in_correct_path(self):
        path = TreePath("home")
        self.assertEqual(str(path), "/home")
        self.assertEqual(path.path_elems, [TreePathElem("home")])

    def test_appending_path_to_path_must_result_in_correct_path(self):
        path = TreePath() / "home" / "kevin"
        self.assertEqual(str(path), "/home/kevin")
        self.assertEqual(path.path_elems, [TreePathElem("home"), TreePathElem("kevin")])

    def test_appending_path_by_division_operator_must_result_in_correct_path(self):
        path = TreePath("/") / "home" / "kevin"
        self.assertEqual(str(path), "/home/kevin")
        self.assertEqual(path.path_elems, [TreePathElem("home"), TreePathElem("kevin")])

    def build_testtree(self):
        builder = TreeBuilder()
        builder.add_path(TreePath("/home/kevin/Documents"))
        builder.add_path(TreePath() / "usr" / "share" / "uae")
        builder.add_path(TreePath("/home/kevin/opt/amiga/uae.info"))
        return builder.root

    def test_treebuilder(self):
        root_node = self.build_testtree()
        self.assertEqual(root_node.name, "<root>")
        self.assertEqual(len(root_node.children), 2)
        home_node = root_node.children[0]
        self.assertEqual(home_node.name, "home")
        self.assertEqual(len(home_node.children), 1)
        kevin_node = home_node.children[0]
        self.assertEqual(kevin_node.name, "kevin")
        self.assertEqual(len(kevin_node.children), 2)
        documents_node = kevin_node.children[0]
        self.assertEqual(documents_node.name, "Documents")
        self.assertEqual(len(documents_node.children), 0)
        opt_node = kevin_node.children[1]
        self.assertEqual(opt_node.name, "opt")
        self.assertEqual(len(opt_node.children), 1)
        amiga_node = opt_node.children[0]
        self.assertEqual(amiga_node.name, "amiga")
        self.assertEqual(len(amiga_node.children), 1)
        uae_node = amiga_node.children[0]
        self.assertEqual(uae_node.name, "uae.info")
        self.assertEqual(len(uae_node.children), 0)
        usr_node = root_node.children[1]
        self.assertEqual(usr_node.name, "usr")
        self.assertEqual(len(usr_node.children), 1)
        share_node = usr_node.children[0]
        self.assertEqual(share_node.name, "share")
        self.assertEqual(len(share_node.children), 1)
        uae_node = share_node.children[0]
        self.assertEqual(uae_node.name, "uae")
        self.assertEqual(len(uae_node.children), 0)

    def test_graphvizitor_must_yield_tree_as_dot(self):
        root_node = self.build_testtree()

        graphvizitor = GraphVizitor()
        root_node.visit_subtree(graphvizitor)

        print(graphvizitor.dot)
        # assert that it is
        self.assertEqual(
            str(graphvizitor.dot).strip(),
            """digraph {
	0 [label=<root>]
	0 -> 1
	0 -> 4
	1 [label=home]
	1 -> 2
	2 [label=kevin]
	2 -> 3
	2 -> 7
	3 [label=Documents]
	7 [label=opt]
	7 -> 8
	8 [label=amiga]
	8 -> 9
	9 [label="uae.info"]
	4 [label=usr]
	4 -> 5
	5 [label=share]
	5 -> 6
	6 [label=uae]
}""".strip())
        # graphvizitor.dot.render('test_graph.gv', format='jpg',view=True)




from functools import partial

import matplotlib

from ratisbona_utils.colors.simple_color import (
    lab_to_rgb,
    cielch_to_lab,
    hsv_to_rgb,
    normalized_to_rgb,
    bounded_or_bust,
)


def cielch_based_rgb_palette(hue_steps: int, lightness, chroma):
    return [
        lab_to_rgb(cielch_to_lab([lightness, chroma, float(i / 10)]))
        for i in range(0, 3600, round(3600 / (hue_steps - 1)))
    ]


def hsv_based_rgb_palette(hue_steps: int, saturation, value):
    return [
        normalized_to_rgb(hsv_to_rgb((float(i / 10), saturation, value)))
        for i in range(0, 3600, round(3600 / (hue_steps - 1)))
    ]


def matplotlib_colormap_to_palette(hue_steps: int, colormap_name: str):
    colormap = matplotlib.colormaps[colormap_name].reversed()
    palette1 = colormap([x / (hue_steps - 1) for x in range(hue_steps)])
    return [normalized_to_rgb((r, g, b)) for r, g, b, a in palette1]


terrain = partial(matplotlib_colormap_to_palette, colormap_name="terrain")
viridis = partial(matplotlib_colormap_to_palette, colormap_name="viridis")


def ega_num_to_rgb(ega: int):
    bounded_or_bust(ega, 0, 63)
    red = 0x55 * (((ega >> 1) & 2) | (ega >> 5) & 1)
    green = 0x55 * ((ega & 2) | (ega >> 4) & 1)
    blue = 0x55 * (((ega << 1) & 2) | (ega >> 3) & 1)
    return red, green, blue


def cga_num_to_normalized_rgb(cga: int):
    bounded_or_bust(cga, 0, 15)
    red = 2 / 3 * (cga & 4) / 4 + 1 / 3 * (cga & 8) / 8
    green = 2 / 3 * (cga & 2) / 2 + 1 / 3 * (cga & 8) / 8
    blue = 2 / 3 * (cga & 1) / 1 + 1 / 3 * (cga & 8) / 8
    if cga == 6:
        green = green * 2 / 3
    return red, green, blue



from .quotes_parser import QuotesParser
from .templates import open_dialogue_document, close_dialogue_document
from .quoting_and_escaping import latex_quote, replace_emojis

__all__ = [
    'QuotesParser',
    'open_dialogue_document',
    'close_dialogue_document',
    'latex_quote',
    'replace_emojis'
]


def open_dialogue_document(title: str) -> str:
    return (
"""\\documentclass[a4paper]{article}
\\usepackage[ngerman]{babel}
\\usepackage[dvipsnames]{xcolor}
\\usepackage[export]{adjustbox} % loads also graphicx
\\usepackage{amsmath}
\\usepackage{emoji}
\\usepackage{fontspec}
\\usepackage{dialogue}
\\usepackage{wasysym}
\\usepackage{minted}
\\usepackage{cprotect}
\\usepackage[colorlinks=true, allcolors=TealBlue]{hyperref}
\\usepackage{longtable,tabu}
\\setlength{\\parindent}{0ex}
\\setlength{\\parskip}{1ex}
\\setminted{breaklines=true, breakanywhere=true}

\\newcommand{\\uliji}[1]{
  {\\setmainfont{Noto Color Emoji}[Renderer=Harfbuzz]{#1}}
}


\\author{Ulrich Schwenk}
\\title{""" + title +"""}

\\begin{document}
\\setlength{\\tabulinesep}{6pt}
\\maketitle
\\newpage
"""
    )

def close_dialogue_document() -> str:
    return "\n\\end{document}\n"


import unittest

from ratisbona_utils.colors.simple_color import f_inv, f


class MyTestCase(unittest.TestCase):

    def test_finv_must_be_inv_of_f(self):
        t=0.0
        while t<100.0:
            fval = f(t)
            finvval = f_inv(fval)
            print(t, fval, finvval)
            self.assertAlmostEqual( t, finvval)
            t += 0.01


if __name__ == '__main__':
    unittest.main()
